from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class NestInit_Args_Implicit:
    """
    GOAL
    ----
    base for using classes with implicit args as variants

    SPECIALLY CREATED FOR
    ---------------------
    EqArgs for cmp all args between each other!
    """
    ARGS: TYPING.ARGS_FINAL

    def __init__(self, *args, **kwargs) -> None:
        self.ARGS = args
        super().__init__(**kwargs)


class NestInit_Kwargs_Implicit:
    """
    GOAL
    ----
    just extend same logic variant
    """
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, *args, **kwargs) -> None:
        self.KWARGS = kwargs
        super().__init__(*args)


class NestInit_ArgsKwargs_Implicit:
    """
    GOAL
    ----
    just extend same logic variant
    """
    ARGS: TYPING.ARGS_FINAL
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs


# =====================================================================================================================



import pathlib
from base_aux.path2_file.m1_filepath import *


# =====================================================================================================================
CWD = pathlib.Path().cwd()


# =====================================================================================================================
class Test_Filepath:
    def test__1_name1_single(self):
        victim = Resolve_FilePath(name="name")
        assert victim.name_PREFIX == ""
        assert victim.name_SUFFIX == ""
        assert victim.NAME == "name"
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == "name"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        # filepath ----
        victim = Resolve_FilePath(name="name2", filepath=CWD.joinpath("name.extlast"))
        assert victim.NAME == "name2"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name2.extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath("name2.extlast")

    def test__1_name2_several(self):
        victim = Resolve_FilePath(name="name1.name2", extlast="extlast")
        assert victim.NAME == "name1.name2"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name1.name2.extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        # dots -----
        victim = Resolve_FilePath(name="name1.name2..", extlast="extlast")
        assert victim.NAME == "name1.name2.."
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name1.name2...extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        # dots -----
        victim = Resolve_FilePath(namefull="name1.name2...extlast", )
        assert victim.NAME == "name1.name2.."
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name1.name2...extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

    def test__2_extlast(self):
        victim = Resolve_FilePath(extlast="extlast")
        assert victim.NAME == ""
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == ".extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(name="name", extlast="extlast")
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name.extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        # filepath ----
        victim = Resolve_FilePath(extlast="extlast2", filepath=CWD.joinpath("name.extlast"))
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast2"
        assert victim.NAMEFULL == "name.extlast2"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath("name.extlast2")

    def test__3_nameext(self):
        victim = Resolve_FilePath(namefull="name.extlast")
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name.extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(namefull="name")
        assert victim.NAME == "name"
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == "name"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(namefull="name.")
        assert victim.NAME == "name"
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == "name."
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(namefull=".extlast")
        assert victim.NAME == ""
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == ".extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        # filepath ----
        victim = Resolve_FilePath(namefull="name2.extlast2", filepath=CWD.joinpath("name.extlast"))
        assert victim.NAME == "name2"
        assert victim.EXTLAST == "extlast2"
        assert victim.NAMEFULL == "name2.extlast2"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath("name2.extlast2")

    def test__4_dirpath(self):
        victim = Resolve_FilePath(dirpath=CWD.joinpath("path2"))
        assert victim.NAME == ""
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == ""
        assert victim.DIRPATH == CWD.joinpath("path2")
        assert victim.FILEPATH == CWD.joinpath("path2")

        victim = Resolve_FilePath(dirpath=CWD.joinpath("path2"), filepath=CWD.joinpath("path1", "name.extlast"))
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name.extlast"
        assert victim.DIRPATH == CWD.joinpath("path2")
        assert victim.FILEPATH == CWD.joinpath("path2", "name.extlast")

    def test__5_filepath(self):
        victim = Resolve_FilePath(filepath=CWD.joinpath("name"))
        assert victim.NAME == "name"
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == "name"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(filepath=CWD.joinpath("name."))
        assert victim.NAME == "name"
        assert victim.EXTLAST == ""
        assert victim.NAMEFULL == "name."
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(filepath=CWD.joinpath(".extlast"))
        assert victim.NAME == ""
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == ".extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(filepath=CWD.joinpath("name.extlast"))
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name.extlast"
        assert victim.DIRPATH == CWD
        assert victim.FILEPATH == CWD.joinpath(victim.NAMEFULL)

        victim = Resolve_FilePath(filepath=CWD.joinpath("path1", "name.extlast"))
        assert victim.NAME == "name"
        assert victim.EXTLAST == "extlast"
        assert victim.NAMEFULL == "name.extlast"
        assert victim.DIRPATH == CWD.joinpath("path1")
        assert victim.FILEPATH == CWD.joinpath("path1", "name.extlast")

    def test__6_all(self):
        victim = Resolve_FilePath(
            prefix="prefix",
            suffix="suffix",
            name="name3",
            extlast="extlast3",
            namefull="name2.extlast2",
            dirpath=CWD.joinpath("path2"),
            filepath=CWD.joinpath("name.extlast")
        )
        assert victim.name_PREFIX == "prefix"
        assert victim.name_SUFFIX == "suffix"

        assert victim.NAME == "name3"
        assert victim.EXTLAST == "extlast3"
        assert victim.NAMEFULL == "prefixname3suffix.extlast3"
        assert victim.DIRPATH == CWD.joinpath("path2")
        assert victim.FILEPATH == CWD.joinpath("path2", "prefixname3suffix.extlast3")


# =====================================================================================================================



class Template_Multiton:
    """
    NOTE
    ----
    never use nesting!
    its only for sample template!
    """
    _MULTITON = {}
    param1: int
    param2: str

    def __new__(cls, param1: int, param2: str, *args, **kwargs):
        key = (param1, param2)
        if key not in cls._MULTITON:
            instance = super().__new__(cls)      # kwArgs NOT NEED!!!
            cls._MULTITON[key] = instance
        return cls._MULTITON[key]

    def __init__(self, param1: int, param2: str, p3=None):
        self.param1 = param1
        self.param2 = param2


obj1 = Template_Multiton(1, "X")
obj2 = Template_Multiton(2, "Y")
obj3 = Template_Multiton(1, "X")  # Возвращает существующий экземпляр
obj4 = Template_Multiton(1, "zzz", 2)

print(f"{obj1.param1=}")
print(f"{obj2.param1=}")

print(obj1 is obj3)  # True



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.numbers.m2_num_single_parsed import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        # TRASH ----
        (None, (None, None, None, )),
        (True, (None, None, None, )),
        (False, (None, None, None, )),
        ("", (None, None, None, )),

        ("a1.2.3a", (None, None, None,)),

        # int ----
        (123, (123, 123, None, )),
        ("123", (123, 123, None, )),
        ("a123a", (123, 123, None, )),

        # float ----
        ("a1.00a", (1.0, None, 1.0, )),
    ]
)
def test__NumParsedSingle(source, _EXPECTED):
    Lambda(NumParsedSingle(source).resolve).check_expected__assert(_EXPECTED[0])
    Lambda(NumParsedSingleInt(source).resolve).check_expected__assert(_EXPECTED[1])
    Lambda(NumParsedSingleFloat(source).resolve).check_expected__assert(_EXPECTED[2])


# =====================================================================================================================



from typing import *

from aiohttp import web
from fastapi import FastAPI, Response
from fastapi.responses import RedirectResponse

from base_aux.servers.m2_server1_aiohttp import *
from base_aux.servers.m3_server2_fastapi import *

from .models import *


# =====================================================================================================================
class TpApi_Aiohttp(ServerAiohttpBase):
    async def response_get_html__(self, request) -> web.Response:
        # --------------------------
        html_block = self.html_block__api_index()

        # RESPONSE --------------------------------------------------
        page_name = "API_INDEX"
        html = self.html_create(data=html_block, redirect_time=2, request=request)
        return web.Response(text=html, content_type='text/html')

    @decorator__log_request_response
    async def response_post__start(self, request) -> web.Response:
        # return self.response_get__start(request)  # this is will not work!
        self.data.signal__tp_start.emit()

        # RESPONSE --------------------------------------------------
        response = web.json_response(data={})
        return response

    @decorator__log_request_response
    async def response_post__stop(self, request) -> web.Response:
        self.data.signal__tp_stop.emit()

        # RESPONSE --------------------------------------------------
        response = web.json_response(data={})
        return response

    @decorator__log_request_response
    async def response_post___reset_duts_sn(self, request) -> web.Response:
        self.data._signal__tp_reset_duts_sn.emit()

        # RESPONSE --------------------------------------------------
        response = web.json_response(data={})
        return response

    # ---------------------------------------------------------
    @decorator__log_request_response
    async def response_get_json__info(self, request) -> web.Response:
        body: dict = self.data.STAND.stand__get_info__general_tcsc()      #FIXME:ADDMODEL???
        response = web.json_response(data=body)
        return response

    @decorator__log_request_response
    async def response_get_json__results(self, request) -> web.Response:
        # RESPONSE --------------------------------------------------
        body: dict = self.data.STAND.tci__get_result()   #FIXME:ADDMODEL???
        return web.json_response(data=body)


# =====================================================================================================================
def create_app__FastApi_Tp(self=None, data: Any = None) -> FastAPI:
    # UNIVERSAL ------------------------------------------------------
    app = FastAPI()
    app.data = data

    # WORK -----------------------------------------------------------
    @app.get("/")
    async def redirect() -> Response:
        return RedirectResponse(url="/docs")

    @app.post("/start")
    async def start() -> bool:
        print(f"access async start")
        # print(f"1=EMIT")
        # app.data.signal__tp_start.emit()
        print(f"2=DIRECT START!")
        app.data.start()
        return True

    @app.post("/stop")
    async def stop() -> bool:
        app.data.signal__tp_stop.emit()
        return True

    @app.get("/info")
    async def info() -> dict:   # -> Model_TpInfo:
        # return Model_TpInfo(**app.data.STAND.stand__get_info__general_tcsc())
        return dict(**app.data.STAND.stand__get_info__general_tcsc())

    @app.get("/results")
    async def results() -> dict:    # -> Model_TpResults:
        # return Model_TpResults(**app.data.STAND.stand__get_results())
        result = dict(**app.data.STAND.stand__get_results())
        return result

    return app


# =====================================================================================================================
class TpApi_FastApi(ServerFastApi_Thread):
    create_app = create_app__FastApi_Tp


# =====================================================================================================================



from base_aux.base_nest_dunders.m4_gsai_annots import *


# =====================================================================================================================
class NestIter_AnnotValues(NestGAI_AnnotAttrIC):
    """
    GOAL
    ----
    iterate annot defined values in position order(nesting is available)

    USAGE
    -----
        class Item:
            pass

        class Example(NestIter_AnnotValues):
            def meth(self):
                pass

            VALUE1: Item = Item(1)
            VALUE3: Item = Item(3)
            VALUE2: Item = Item(2)
            VALUE20: Item
            VALUE200 = 200

        for i in Example():
            print(i)

        ---> Item(1), Item(3), Item(2)

    SPECIALLY CREATED FOR
    ---------------------
    pyqt.pte_highlights.StylesPython

    WHY NOT - NAMEDTUPLE
    --------------------
    typing.NamedTuple is good as collection! but nesting is not accepted!
    """

    def __iter__(self):
        yield from AttrAux_AnnotsAll(self).iter__annot_values()


# =====================================================================================================================



from typing import *
import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_dict.m2_dict_ic import *
from base_aux.aux_dict.m3_dict_ga import *


# =====================================================================================================================
@pytest.mark.parametrize(argnames="VictimCls", argvalues=[DictIcKeys, DictIcKeys_Ga])
def test__universe(VictimCls):
    victim1 = VictimCls()
    victim1['NAme'] = 'VALUE'
    victim1[1] = 1

    victim2 = VictimCls(NAme='VALUE')
    victim2[1] = 1

    victim3 = VictimCls({"NAme": 'VALUE', 1: 1})

    # -------------------------------------------
    for victim in [victim1, victim2, victim3, ]:
        assert len(victim) == 2

        assert list(victim) == ["NAme", 1]

        # EQ
        assert victim == {1: 1, "naME": 'VALUE'}
        assert victim == {"naME": 'VALUE', 1: 1}

        assert victim != {"naME": 'VALUE', 1: 11}
        assert victim != {"naME": 'VALUE', 11: 1}
        assert victim != {"naME": 'VALUE2', 1: 1}
        assert victim != {"naME2": 'VALUE', 1: 1}

        # CONTAIN
        assert 'naME' in victim
        assert 1 in victim
        assert 0 not in victim

        # ACCESS
        Lambda(lambda: victim[0]).check_expected__assert(Exception)   # keep original behaviour - maybe need switch to None???
        Lambda(lambda: victim.get(0)).check_expected__assert(None)

        assert victim[1] == 1
        assert victim.get(1) == 1
        assert victim['name'] == "VALUE"
        assert victim['NAME'] == "VALUE"
        assert victim.get('NAME') == "VALUE"

        # update
        len0 = len(victim)
        victim['name'] = 'VALUE2'
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE2"

        victim['NAME'] = 'VALUE3'
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE3"

        victim.update({'NaMe': 'VALUE4'})
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE4"

        assert list(victim) == ["NAme", 1]

        # del ------
        try:
            del victim['name222']
        except:
            assert True
        else:
            assert False

        del victim['name']
        assert list(victim) == [1, ]
        victim['NAme'] = 'VALUE'

        # pop ------
        try:
            victim.pop('name222')
        except:
            assert True
        else:
            assert False

        assert victim.pop('name') == "VALUE"
        assert list(victim) == [1, ]
        # victim['NAme'] = 'VALUE'


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, keys_all_str, _EXPECTED",
    argvalues=[
        (dict(), True, []),
        (dict(attr1=1), True, ["attr1", ]),
        (dict(ATTR1=1), True, ["ATTR1", ]),
        ({1:1}, False, [1, ]),
    ]
)
@pytest.mark.parametrize(argnames="VictimClsPair", argvalues=[(DictIcKeys, DictIc_LockedKeys), (DictIcKeys_Ga, DictIc_LockedKeys_Ga)])
def test__keys(VictimClsPair, source, keys_all_str, _EXPECTED):
    Lambda(list(VictimClsPair[0](source))).check_expected__assert(_EXPECTED)
    Lambda(list(VictimClsPair[1](source))).check_expected__assert(_EXPECTED)

    if not keys_all_str:
        return

    Lambda(list(VictimClsPair[0](**source))).check_expected__assert(_EXPECTED)
    Lambda(list(VictimClsPair[1](**source))).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, key, keys_all_str, _EXPECTED",
    argvalues=[
        (dict(), "attr1", True, None),
        (dict(attr1=1), "attr1", True, "attr1"),
        (dict(attr1=1), "ATTR1", True, "attr1"),
        (dict(ATTR1=1), "ATTR1", True, "ATTR1"),
        (dict(attr1=1), "hello", True, None),

        (dict(attr1=1), 0, True, None),
        ({1:1}, 0, False, None),
        ({1:1}, 1, False, 1),
    ]
)
@pytest.mark.parametrize(argnames="VictimClsPair", argvalues=[(DictIcKeys, DictIc_LockedKeys), (DictIcKeys_Ga, DictIc_LockedKeys_Ga)])
def test__key__get_original(VictimClsPair, source, key, keys_all_str, _EXPECTED):
    Lambda(VictimClsPair[0](source).key__get_original(key)).check_expected__assert(_EXPECTED)
    Lambda(VictimClsPair[1](source).key__get_original(key)).check_expected__assert(_EXPECTED)

    if not keys_all_str:
        return

    Lambda(VictimClsPair[0](**source).key__get_original(key)).check_expected__assert(_EXPECTED)
    Lambda(VictimClsPair[1](**source).key__get_original(key)).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, key, keys_all_str, _EXPECTED",
    argvalues=[
        (dict(attr1=1), "attr1", True, [None, None]),
        (dict(attr1=1), "ATTR1", True, [None, None]),
        (dict(ATTR1=1), "ATTR1", True, [None, None]),
        (dict(attr1=1), "hello", True, [None, Exception]),

        (dict(attr1=1), 0, False, [None, Exception]),
        ({1:1}, 0, False, [None, Exception]),
        ({1:1}, 1, False, [None, None]),
    ]
)
@pytest.mark.parametrize(argnames="VictimClsPair", argvalues=[(DictIcKeys, DictIc_LockedKeys), (DictIcKeys_Ga, DictIc_LockedKeys_Ga)])
def test__si_update(VictimClsPair, source, key, keys_all_str, _EXPECTED):
    # -------------------------------------------------
    victim = VictimClsPair[0](source)
    Lambda(victim.update({key: 11})).check_expected__assert(_EXPECTED[0])
    Lambda(victim.get(key)).check_expected__assert(11)

    victim[key] = 111
    Lambda(victim.get(key)).check_expected__assert(111)

    if keys_all_str:
        Lambda(victim.update(**{key: 1111})).check_expected__assert(_EXPECTED[0])
        Lambda(victim.get(key)).check_expected__assert(1111)

        # kwargs out
        Lambda(dict(**VictimClsPair[0](source))).check_expected__assert(source)
        Lambda(dict(**VictimClsPair[1](source))).check_expected__assert(source)

    # -------------------------------------------------
    victim = VictimClsPair[1](source)
    Lambda(lambda: victim.update({key: 2})).check_expected__assert(_EXPECTED[1])

    if _EXPECTED[1] == Exception:
        assert key not in victim

        # ---------------
        try:
            victim[key] = 123
        except:
            pass
        else:
            assert False
        Lambda(victim.get(key)).check_expected__assert(None)

        # ---------------
        try:
            victim.update({key: 123})
        except:
            pass
        else:
            assert False
        Lambda(victim.get(key)).check_expected__assert(None)

    else:
        assert key in victim

        # ---------------
        Lambda(victim.update({key: 11})).check_expected__assert(_EXPECTED[1])
        Lambda(victim.get(key)).check_expected__assert(11)

        victim[key] = 111
        Lambda(victim.get(key)).check_expected__assert(111)

    # if not keys_all_str:
    #     return
    #
    # Lambda(VictimClsPair[0](**source).key__get_original(key)).expect__check_assert(_EXPECTED)
    # Lambda(VictimClsPair[1](**source).key__get_original(key)).expect__check_assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.skip
@pytest.mark.parametrize(argnames="VictimCls",argvalues=[DictIcKeys, DictIc_LockedKeys, DictIcKeys_Ga, DictIc_LockedKeys_Ga])
@pytest.mark.parametrize(
    argnames="source",
    argvalues=[
        dict(),
        dict(a1=1),
    ]
)
def test__kwargs_unpack(VictimCls, source):
    Lambda(dict(**VictimCls({key.lower(): value for key, value in source.items()}))).check_expected__assert({key.upper(): value for key, value in source.items()})


# =====================================================================================================================



import numpy as np
import pandas as pd

from dataclasses import dataclass
from base_aux.aux_np_pd.m0_typing import *

from base_aux.base_nest_dunders.m1_init3_params_dict_kwargs_update import *
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import *
from base_aux.aux_dict.m2_dict_ic import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *

# calculates ta_meth!!!
import pandas_ta as ta  # VERY IMPORTANT!!! even if no used!


# =====================================================================================================================
class ColumnSettings(NamedTuple):
    EQ: str | Base_EqValid | None = None    # none is when you dont know exact
    ROUND: int = 1


# =====================================================================================================================
class Base_Indicator(NestInit_Source, NestInit_ParamsDict_UpdateByKwargs):
    """
    GOAL
    ----
    ON INIT
        1/ init source as PD - dont do smth with source - no shrink! only add new calculations!
        2/ calculate additional exact Indicator PD
    ACCESS
        3/ access to indicator values

    SPECIALLY CREATED FOR
    ---------------------
    input data from mt5
    calculate indicator values
    """
    SOURCE: TYPING__NP_TS__FINAL    # HISTORY input - from MT5 - todo: update to DF? use with loaded TA data?
    DF: TYPING__PD_DATAFRAME        # result OUTPUT - from PD_TA

    # ---------------------------
    NAME: str = "DEF_IndNameInfo"                       # just info!
    PARAMS: DictIc_LockedKeys_Ga
    COLUMN_SETINGS: DictIcKeys[str, ColumnSettings]     # if not know what to use - keep blanc str "" or None!!!

    @property
    def TA_METH(self) -> Callable[..., TYPING__PD_DATAFRAME | TYPING__PD_SERIES]:
        """
        GOAL
        ----
        exact method for making calculations for TA

        NOTE
        ----
        use property! or find solution to use in classmethod!!
        """
        raise NotImplementedError()

    @property
    def HISTORY_ENOUGH_THRESH(self) -> int:
        """
        GOAL
        ----
        calculate (for exact params) history depth wich is enough for correct calculations the indicator

        ORIGINAL IDEA
        -------------
        РАСЧЕТ ДЛИНЫ БАРОВ
            количество влияет на результат!!!!
            при не особо достаточном количестве баров - расчет произойдет НО значения будут отличаться от фактического!!!
            видимо из-за того что будет расчитываться с нулевыми некоторыми начальными значениями!!!

            sum * 2 = это очень мало!!!!!
            sum * 10 = кажется первая, что вообще показала полное совпадение с Tinkoff терминалом!!!

            ADX
                !!! ЭТО ОЧЕНЬ ВАЖНО ДЛЯ ADX !!!!
            STOCH
                вообще не важно - кажется там сколько длина его - столько и баров достаточно!!!
        """
        return sum(self.PARAMS.values()) * 10

    # -----------------------------------------------------------------------------------------------------------------
    # TODO: decide what to do with Series/Tail/Last or use help to direct access after!!! --- use originally indexing

    def __getattr__(self, item: str) -> TYPING__PD_SERIES | NoReturn:
        """
        GOAL
        ----
        return exact Indicator values series from DF!!!

        NOTE
        ----
        result_last_element = df.iloc[len(df) - 1]
        result_full_series = df
        result_tail_cut = df[-return_tail::]
        """
        try:
            column_name = self.COLUMN_SETINGS.key__get_original(item)
            return self.DF[column_name]
        except Exception as exc:
            Warn(f"{item=}/{exc!r}")
            raise exc

    # -----------------------------------------------------------------------------------------------------------------
    def _init_post(self) -> None:
        """
        GOAL
        ----
        do all final calculates and fixes on source
        """
        self._init_post0__pd_set_options()
        self._init_post0__fix_attrs()
        self._init_post1__warn_if_not_enough_history()
        self._init_post2__calculate_ta()
        self._init_post3__ensure_df()
        # self._init_post4__apply_time_indexes()
        self._init_post5__rename_columns()
        self._init_post6__round_values()
        self.init_post7__calculate_extra_columns()

    @staticmethod
    def _init_post0__pd_set_options():
        pd.set_option('display.max_columns', 500)  # сколько столбцов показываем
        pd.set_option('display.width', 1500)  # макс. ширина таблицы для показа

    def _init_post0__fix_attrs(self) -> None:
        """
        GOAL
        ----
        fix/remake/create all attrs if need
        """
        self.DF = pd.DataFrame(self.SOURCE)
        self.COLUMN_SETINGS = DictIcKeys(self.COLUMN_SETINGS)    # just make a cls COPY to self

    def _init_post1__warn_if_not_enough_history(self) -> None:
        """
        GOAL
        ----
        main goal - Warn if not enough lines to calculate correct values
        """
        len_source = len(self.SOURCE)
        try:
            if len_source < self.HISTORY_ENOUGH_THRESH:
                Warn(f"{len_source=}/{self.HISTORY_ENOUGH_THRESH=}")
        except Exception as exc:
            Warn(f"{len_source=}/{exc!r}")

    def _init_post2__calculate_ta(self) -> None:
        """
        GOAL
        ----
        do exact TA calculations
        """
        self.DF = self.TA_METH(**self.PARAMS)

    def _init_post3__ensure_df(self) -> None:
        """
        GOAL
        ----
        when indicator calculated into pdSeries instead of pdDataframe
        like for singleDimentional as Wma/Stoch/
        1. reformat into DF
        2. set name if NONAME column
        """
        if isinstance(self.DF, pd.core.series.Series):
            self.DF = pd.DataFrame(self.SOURCE)

    def _init_post4__apply_time_indexes(self) -> None:
        """
        GOAL
        ----
        pd_ta create indexes only as position index!
        so i want to apply original "time" indexes and be able to combine/concatenate several DF expecting same TIME!
        at least for cmp!
        """
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH
        # TODO: FINISH

    def _init_post5__rename_columns(self) -> None:
        """
        GOAL
        ----
        rename columns to use finals simple names!
        """
        for col_original in self.DF.columns:
            for col_name, col_settings in self.COLUMN_SETINGS.items():
                if col_settings.EQ == col_original:
                    self.DF.rename({col_original: col_name})

    def _init_post6__round_values(self) -> None:
        """
        GOAL
        ----
        round indicator calculations
        """
        # df = df.iloc[:].round(indicator_obj.ROUND)    # old logic

        for col_name, col_settings in self.COLUMN_SETINGS.items():
            # round - did not work INLINE!!! - need resave!
            self.DF[col_name] = self.DF[col_name].round(col_settings.ROUND)

    def init_post7__calculate_extra_columns(self) -> None:
        """
        GOAl
        ----
        do extra calculations like for geom sums in addition for common indicator
        """
        return NotImplemented()


# =====================================================================================================================



from base_aux.versions.m2_version import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m3_exceptions import *

from base_aux.tables.m1_table_obj import TableObj


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="schema, _EXP_valid, _EXP_count_lines, _EXP_count_col",
    argvalues=[
        # CORRECT -------------
        (dict(), True, 0, 0),
        (dict(ATC=[]), True, 1, 0),
        (dict(ATC=[1,]), True, 1, 1),
        (dict(ATC=[1,2,]), True, 1, 2),
        (dict(ATC=[1,], PTB=[2,]), True, 2, 1),
        (dict(ATC=[1,], PTB=(2,)), True, 2, 1),
        (dict(ATC=[1,], PTB={2,}), True, 2, 1),
        (dict(ATC=[], PTB=[]), True, 2, 0),

        (dict(ATC="atc", PTB=[1, 2, 3]), True, 2, 3),
        (dict(ATC="atc", PTB={1:1, 2:2, 3:3}), True, 2, 3),

        # INCORRECT -----------
        (dict(ATC=1), False, None, None),
        (dict(ATC=[1,], PTB=[]), False, None, None),
        (dict(ATC=[], PTB=0), False, None, None),
        (1, False, None, None),
        ({1:1}, False, None, None),
    ]
)
def test__full(schema, _EXP_valid, _EXP_count_lines, _EXP_count_col):
    Lambda(TableObj._validate_schema, schema=schema).check_expected__assert(_EXP_valid)
    Lambda(lambda: TableObj(**schema)).check_raised__assert(not _EXP_valid)

    if _EXP_valid:
        Lambda(TableObj(**schema).schema).check_expected__assert(schema)
        Lambda(TableObj(**schema).count_lines).check_expected__assert(_EXP_count_lines)
        Lambda(TableObj(**schema).count_columns).check_expected__assert(_EXP_count_col)

        for name, value in schema.items():
            Lambda(lambda: getattr(TableObj(**schema), name)).check_expected__assert(value)
            Lambda(lambda: TableObj(**schema)[name]).check_expected__assert(value)

        Lambda(lambda: getattr(TableObj(**schema), "NONAME")).check_expected__assert(Exception)
        Lambda(lambda: TableObj(**schema)["NONAME"]).check_expected__assert(Exception)


# =====================================================================================================================



import pandas as pd
from base_aux.base_types.m2_info import ObjectInfo


series_a = pd.DataFrame([1, 2, 3])
print(type(series_a))   # <class 'pandas.core.series.Series'>
# ObjectInfo(series_a).print()

# print(series_a)
"""
a    NaN
b    6.0
c    8.0
d    NaN
dtype: float64
"""



import pytest

from base_aux.aux_argskwargs.m4_kwargs_eq_expect import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *
from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m3_exceptions import *
from base_aux.base_values.m4_primitives import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="other_draft, eq_kwargs, eq_expects, _EXP_checkIf, _EXP_ga",
    argvalues=[
        # SINGLE =============================================
        # no expected -------
        (1, dict(eq1=1), dict(), [True, True, False, False], [True, Exception]),
        (1, dict(eq1=EqValid_EQ(1)), dict(), [True, True, False, False], [True, Exception]),
        (1, dict(eq1=11), dict(), [False, False, True, True], [False, Exception]),
        (1, dict(eq1=EqValid_EQ(11)), dict(), [False, False, True, True], [False, Exception]),
        (lambda: 1, dict(eq1=1), dict(), [True, True, False, False], [True, Exception]),

        # expected -------
        (1, dict(eq1=1), dict(eq1=True), [True, True, False, False], [True, Exception]),
        (1, dict(eq1=1), dict(eq1=False), [False, False, True, True], [True, Exception]),

        # RAISED --------
        (1, dict(eq1=1), dict(eq111=True), [Exc__WrongUsage, Exc__WrongUsage, Exc__WrongUsage, Exc__WrongUsage], [True, Exception]),

        (1, dict(eq1=ClsEqRaise()), dict(), [False, False, True, True], [False, Exception]),                # NOTE: this is not correct usage! nobody will not do this!
        (1, dict(eq1=ClsEqRaise()), dict(eq1=Exception), [True, True, False, False], [False, Exception]),   # NOTE: this is not correct usage! nobody will not do this!
        (LAMBDA_RAISE, dict(eq1=1), dict(), [False, False, True, True], [False, Exception]),
        (LAMBDA_RAISE, dict(eq1=1), dict(eq1=Exception), [False, False, True, True], [False, Exception]),           # FIXME: CAREFUL!!! INCORRECT!!!
        (LAMBDA_RAISE, dict(eq1=Exception), dict(eq1=Exception), [False, False, True, True], [False, Exception]),   # FIXME: CAREFUL!!! INCORRECT!!!

        # MULTY =============================================
        (1, dict(eq1=1, eq2=11), dict(), [False, True, False, True], [True, False]),
        (1, dict(eq1=1, eq2=11), dict(eq1=True, eq2=True), [False, True, False, True], [True, False]),

        (1, dict(eq1=1, eq2=11), dict(eq1=True), [True, True, False, False], [True, False]),
        (1, dict(eq1=1, eq2=11), dict(eq1=True, eq2=False), [True, True, False, False], [True, False]),
    ]
)
def test__base_eq_kwargs(other_draft, eq_kwargs, eq_expects, _EXP_checkIf, _EXP_ga):
    Victim = Base_KwargsEqExpect
    Lambda(Victim(other_draft, **eq_kwargs).bool_if__all_true, **eq_expects).check_expected__assert(_EXP_checkIf[0])
    Lambda(Victim(other_draft, **eq_kwargs).bool_if__any_true, **eq_expects).check_expected__assert(_EXP_checkIf[1])
    Lambda(Victim(other_draft, **eq_kwargs).bool_if__all_false, **eq_expects).check_expected__assert(_EXP_checkIf[2])
    Lambda(Victim(other_draft, **eq_kwargs).bool_if__any_false, **eq_expects).check_expected__assert(_EXP_checkIf[3])

    Lambda(Victim(other_draft, **eq_kwargs).raise_if__all_true, **eq_expects).check_expected__assert(Exception if _EXP_checkIf[0] else False)
    Lambda(Victim(other_draft, **eq_kwargs).raise_if__any_true, **eq_expects).check_expected__assert(Exception if _EXP_checkIf[1] else False)
    Lambda(Victim(other_draft, **eq_kwargs).raise_if__all_false, **eq_expects).check_expected__assert(Exception if _EXP_checkIf[2] else False)
    Lambda(Victim(other_draft, **eq_kwargs).raise_if__any_false, **eq_expects).check_expected__assert(Exception if _EXP_checkIf[3] else False)

    # GA
    Lambda(lambda: Victim(other_draft, **eq_kwargs).eq1).check_expected__assert(_EXP_ga[0])
    Lambda(lambda: Victim(other_draft, **eq_kwargs).EQ1).check_expected__assert(_EXP_ga[0])
    Lambda(lambda: Victim(other_draft, **eq_kwargs).eq2).check_expected__assert(_EXP_ga[1])
    Lambda(lambda: Victim(other_draft, **eq_kwargs).EQ2).check_expected__assert(_EXP_ga[1])
    Lambda(lambda: Victim(other_draft, **eq_kwargs).HELLO).check_expected__assert(Exception)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="other_draft, eq_expects, _EXP_checkIf, _EXP_ga",
    argvalues=[
        ("linux", dict(linux=True), [True, True, False, False], [True, False]),
        ("LINux", dict(linUX=True), [True, True, False, False], [True, False]),
        (lambda: "LINux", dict(linUX=True), [True, True, False, False], [True, False]),

        ("windows", dict(linux=True), [False, False, True, True], [False, True]),
        ("windows", dict(linux=True, windows=True), [False, True, False, True], [False, True]),
        ("Windows", dict(linux=True, windows=True), [False, True, False, True], [False, True]),

    ]
)
def test__OS(other_draft, eq_expects, _EXP_checkIf, _EXP_ga):
    Victim = KwargsEqExpect_OS
    Lambda(Victim(other_draft).bool_if__all_true, **eq_expects).check_expected__assert(_EXP_checkIf[0])
    Lambda(Victim(other_draft).bool_if__any_true, **eq_expects).check_expected__assert(_EXP_checkIf[1])
    Lambda(Victim(other_draft).bool_if__all_false, **eq_expects).check_expected__assert(_EXP_checkIf[2])
    Lambda(Victim(other_draft).bool_if__any_false, **eq_expects).check_expected__assert(_EXP_checkIf[3])

    Lambda(Victim(other_draft).raise_if__all_true, **eq_expects).check_expected__assert(Exc__Expected if _EXP_checkIf[0] else False)
    Lambda(Victim(other_draft).raise_if__any_true, **eq_expects).check_expected__assert(Exc__Expected if _EXP_checkIf[1] else False)
    Lambda(Victim(other_draft).raise_if__all_false, **eq_expects).check_expected__assert(Exc__Expected if _EXP_checkIf[2] else False)
    Lambda(Victim(other_draft).raise_if__any_false, **eq_expects).check_expected__assert(Exc__Expected if _EXP_checkIf[3] else False)

    # GA
    Lambda(lambda: Victim(other_draft).linux).check_expected__assert(_EXP_ga[0])
    Lambda(lambda: Victim(other_draft).LINUX).check_expected__assert(_EXP_ga[0])
    Lambda(lambda: Victim(other_draft).windows).check_expected__assert(_EXP_ga[1])
    Lambda(lambda: Victim(other_draft).WINDOWS).check_expected__assert(_EXP_ga[1])
    Lambda(lambda: Victim(other_draft).HELLO).check_expected__assert(Exception)


def test__os_2():
    if platform.system().lower() == "windows":
        assert KwargsEqExpect_OS().bool_if__any_true(windows=True)
    else:
        assert not KwargsEqExpect_OS().bool_if__any_true(windows=True)


# =====================================================================================================================



from base_aux.aux_argskwargs.m2_argskwargs_aux import *
from base_aux.base_types.m0_static_types import *
from base_aux.base_types.m0_static_typing import TYPING


# =====================================================================================================================
class NestInit_SourceKwArgs_Implicit(NestInit_Source):
    """
    GOAL
    ----
    just to make inition source with KwArgs

    SPECIALLY CREATED FOR
    ---------------------
    Lambda.expect__check_assert
    """
    ARGS: TYPING.ARGS_FINAL
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, source: Any = None, *args, **kwargs) -> None:
        self.ARGS = args
        self.KWARGS = kwargs
        super().__init__(source)


# =====================================================================================================================
class _NestInit_SourceKwArgs_Explicite(NestInit_Source):
    """
    NOTE
    ----
    try NOT TO USE IT!!!
    use NestInit_SourceKwArgs_Implicit instead! as clearest!
    """
    ARGS: TYPING.ARGS_FINAL
    KWARGS: TYPING.KWARGS_FINAL

    def __init__(self, source: Any = None, args: TYPING.ARGS_DRAFT = (), kwargs: TYPING.KWARGS_DRAFT = None, *args2, **kwargs2) -> None:
        self.ARGS = ArgsKwargsAux(args).resolve_args()
        self.KWARGS = ArgsKwargsAux(kwargs).resolve_kwargs()
        super().__init__(source, *args2, **kwargs2)


# =====================================================================================================================



import pytest
from typing import *

from base_aux.buses.m1_serial1_client import *
from base_aux.buses.m1_serial3_server import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
@pytest.mark.skip   # FIXME: need to fix not working EMU
class Test__Emulated:
    Victim: type[SerialClient_Emulated]
    victim: SerialClient_Emulated

    @classmethod
    def setup_class(cls):
        class Victim(SerialClient_Emulated):
            pass

        cls.Victim = Victim
        cls.victim = cls.Victim()
        if not cls.victim.connect(_raise=False):
            msg = f"[ERROR] not found PORT PAIRED"
            print(msg)
            raise Exception(msg)

    @classmethod
    def teardown_class(cls):
        if cls.victim:
            cls.victim.disconnect()

    # def setup_method(self, method):
    #     pass
    #     # self.victim.connect(_raise=False)
    #
    # def teardown_method(self, method):
    #     pass
    #     # if self.victim:
    #     #     self.victim.disconnect()

    # -----------------------------------------------------------------------------------------------------------------
    def test__reconnect(self):
        assert self.victim.check__connected() is True
        assert self.victim._EMULATOR__INST.SERIAL_CLIENT.check__connected() is True

        assert self.victim._EMULATOR__INST.isRunning() is True
        assert self.victim.write_read__last("echo 123") == "echo 123"

        self.victim.disconnect()
        assert self.victim.check__connected() is False
        assert self.victim._EMULATOR__INST.SERIAL_CLIENT.check__connected() is False

        assert self.victim._EMULATOR__INST.isRunning() is False

        self.victim.connect()
        assert self.victim.check__connected() is True
        assert self.victim._EMULATOR__INST.SERIAL_CLIENT.check__connected() is True

        assert self.victim._EMULATOR__INST.isRunning() is True
        assert self.victim.write_read__last("echo 123") == "echo 123"

    @pytest.mark.skip
    def test__old(self):
        self.victim.ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__PAIRED
        self.victim._EMULATOR__INST = SerialServer_Base()
        self.victim._EMULATOR__START = True

        assert self.victim._EMULATOR__INST.isRunning() is False
        assert self.victim.connect(_raise=False)
        assert self.victim._EMULATOR__INST.isRunning() is True

        assert self.victim.ADDRESS == self.victim.ADDRESSES__PAIRED[0][0]
        assert self.victim.ADDRESS == self.victim.addresses_paired__get_used()[0]

        assert self.victim._EMULATOR__INST.SERIAL_CLIENT.ADDRESS == self.victim.ADDRESSES__PAIRED[0][1]
        assert self.victim._EMULATOR__INST.SERIAL_CLIENT.ADDRESS == self.victim.addresses_paired__get_used()[1]

        assert self.victim.write_read__last("echo 123") == "echo 123"

        assert self.victim._write("echo 123")
        assert self.victim.read_line(_timeout=1) == "echo 123"

        # RECONNECT ----------------
        self.victim.disconnect()

        assert self.victim._EMULATOR__INST.isRunning() is False
        assert self.victim.connect(_raise=False)
        assert self.victim._EMULATOR__INST.isRunning() is True

        assert self.victim.write_read__last("echo 123") == "echo 123"


# =====================================================================================================================



from typing import *

from base_aux.aux_attr.m1_annot_attr1_aux import AttrAux_Existed


# =====================================================================================================================
class NestStR_AttrsNotPrivate:
    """
    GOAL
    ----
    apply str/repr for show attrs names+values

    CAREFUL
    -------
    dont use in Nest* classes - it can used only in FINALs!!! cause it can have same or meaning is not appropriate!
    """
    def __str__(self):  # it can used only in FINAL!!! dont use in Nest*!!!
        result = ""
        for attr_name in AttrAux_Existed(self).iter__names_filter__not_private():
            if result:
                result += ","
            elif not result:
                # result += f"{self.__class__.__name__}("
                result += f"ATTRS("
            value = getattr(self, attr_name)
            result += f"{attr_name}={value}"

        result += ")"
        return result

    def __repr__(self):       # it can used only in FINAL!!! dont use in Nest*!!!
        return str(self)


# =====================================================================================================================
class NestStR_AttrsNotHidden:
    """
    GOAL
    ----
    apply str/repr for show attrs names+values

    CAREFUL
    -------
    dont use in Nest* classes - it can used only in FINALs!!! cause it can have same or meaning is not appropriate!
    """
    def __str__(self):  # it can used only in FINAL!!! dont use in Nest*!!!
        result = ""
        for attr_name in AttrAux_Existed(self).iter__names_filter__not_hidden():
            if result:
                result += ","
            elif not result:
                # result += f"{self.__class__.__name__}("
                result += f"ATTRS("
            value = getattr(self, attr_name)
            result += f"{attr_name}={value}"

        result += ")"
        return result


# =====================================================================================================================
pass
pass
pass


class NestStR_AttrsPattern:
    """
    GOAL
    ----
    use pattern to make a special string for instance
    """
    NEST_STR__ATTR_PATTERN: str

    def __str__(self):
        return self.NEST_STR__ATTR_PATTERN.format(self)

    def __repr__(self):
        return str(self)


# =====================================================================================================================
def _examples() -> None:
    class Victim(NestStR_AttrsNotPrivate):
        A0: int
        A1: int = 1

    victim = Victim()
    print(victim)


    class Victim(NestStR_AttrsPattern):
        NEST_STR__ATTR_PATTERN = "A0={0.A0}, A1={0.A1}"
        A0: int = 0
        A1: int = 1

    victim = Victim()
    print(victim)


# =====================================================================================================================
if __name__ == "__main__":
    _examples()


# =====================================================================================================================



from base_aux.testplans.tc import TYPING__RESULT_W_NORETURN
from base_aux.valid.m3_valid_chains import ValidChains, Valid
import time

from . import tc1_direct


# =====================================================================================================================
class TestCase(tc1_direct.TestCase):
    ASYNC = True
    DESCRIPTION = "reverse1"

    @classmethod
    def startup__cls__wrapped(cls) -> TYPING__RESULT_W_NORETURN:
        result_chain = ValidChains(
            [
                Valid(value_link=True, name="TRUE"),
                Valid(value_link=False, name="FALSE"),
                Valid(value_link=None, name="NONE"),
            ],
        )
        return result_chain

    def run__wrapped(self) -> bool:
        time.sleep(0.1)
        self.details_update({"detail_value": not self.DEV_COLUMN.DUT.VALUE})
        return not self.DEV_COLUMN.DUT.VALUE


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m7_cmp import *
import operator

from base_aux.base_values.m6_operator import gtlt, gtle, gelt, gele


# =====================================================================================================================
class Victim(NestCmp_GLET_DigitAccuracy):
    def __init__(self, source: float | int, **kwargs):
        self.SOURCE = source
        super().__init__(**kwargs)

    @property
    def CMP_VALUE(self) -> TYPING.DIGIT_FLOAT_INT:
        return self.SOURCE


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, init_acc_vp, run_acc_vp, _EXPECTED",
    argvalues=[
        (10, (None, None), (None, None), (0, 0)),
        (10, (0, None), (None, None), (0, 0)),

        (10, (1, None), (None, None), (1, 1)),
        (10, (None, 1), (None, None), (0.1, 0.1)),
        (10, (None, 1), (1, None), (0.1, 1)),
        (10, (1, None), (None, 1), (1, 0.1)),
    ]
)
def test__cmp_accuracy__last(
        source: float | None,
        init_acc_vp: tuple[float | None, float | None],
        run_acc_vp: tuple[float | None, float | None],
        _EXPECTED: tuple[float, float],
):
    # INIT
    victim = Victim(source, cmp_accuracy_value=init_acc_vp[0], cmp_accuracy_percent=init_acc_vp[1])
    assert victim._cmp_accuracy__last == _EXPECTED[0]

    victim.cmp_ge(2)
    assert victim._cmp_accuracy__last == _EXPECTED[0]

    # call
    victim.cmp_ge(3, *run_acc_vp)
    assert victim._cmp_accuracy__last == _EXPECTED[1]

    # 2=after _cmp_accuracy__get_active
    assert victim._cmp_accuracy__get_active() == _EXPECTED[0]
    assert victim._cmp_accuracy__last == _EXPECTED[0]

    assert victim._cmp_accuracy__get_active(*run_acc_vp) == _EXPECTED[1]
    assert victim._cmp_accuracy__last == _EXPECTED[1]


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="accuracy_v, accuracy_p, _EXPECTED",
    argvalues=[
        (None, None, True),
        (0, None, True),
        (None, 0, True),

        (0, 0, False),
        ("exc", None, False),
        ("exc", 0, False),
        (None, "exc", False),
        (0, "exc", False),
        ("exc", "exc", False),
    ]
)
def test__cmp_accuracy__check_correctness(
        accuracy_v: float | None,
        accuracy_p: float | None,
        _EXPECTED: bool,
):
    victim = Victim(1)

    Lambda(victim._cmp_accuracy__check_correctness, accuracy_v, accuracy_p, False).check_expected__assert(_EXPECTED)

    Lambda(victim._cmp_accuracy__check_correctness, accuracy_v, accuracy_p).check_expected__assert(True if _EXPECTED else Exception)
    Lambda(victim._cmp_accuracy__check_correctness, accuracy_v, accuracy_p, True).check_expected__assert(True if _EXPECTED else Exception)

    Lambda(Victim, 1, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p).check_no_raised__assert(_EXPECTED)
    Lambda(Victim, 1, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p).check_expected__assert(NestCmp_GLET_DigitAccuracy if _EXPECTED else Exception)


# ---------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames="source, accuracy_p, _EXPECTED",
    argvalues=[
        (0, None, 0),
        (0, 0, 0),
        (1, 0, 0),

        (100, 0, 0),
        (100, 0.1, 0.1),
        (100, 1, 1),
        (100, 10, 10),
        (100, 1000, 1000),

        (0.1, 0, 0),
        # (0.1, 0.1, 0.001),  # FIXME! use cmp by percent!
        (0.1, 1, 0.001),
        (0.1, 10, 0.01),
        (0.1, 1000, 1),
    ]
)
def test__cmp_accuracy__translate_from_percent(
        source: float | None,
        accuracy_p: float | None,
        _EXPECTED: float,
):
    victim = Victim(source)

    if _EXPECTED is Exception:
        Lambda(victim._cmp_accuracy__translate_from_percent, accuracy_p).check_expected__assert(_EXPECTED)

    Lambda(victim._cmp_accuracy__translate_from_percent, accuracy_p).check_expected__assert(_EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, other, accuracy_vp, _EXP_GLET, _EXP_EQ",
    argvalues=[
        # TRIVIAL ---------------------------------
        ("exc", 1,  (0, None), (Exception, Exception, Exception, Exception), (Exception, Exception)),
        (1, "exc",  (0, None), (Exception, Exception, Exception, Exception), (Exception, Exception)),
        (1, 1, ("exc", None), (Exception, Exception, Exception, Exception), (Exception, Exception)),
        (1, 1, (None, "exc"), (Exception, Exception, Exception, Exception), (Exception, Exception)),
        (1, 1, (0, 0), (Exception, Exception, Exception, Exception), (Exception, Exception)),

        # 1+1--------------------------------------
        (1, 1,      (None, None), (False, True, True, False), (True, False)),
        (1, 1.0,    (None, None), (False, True, True, False), (True, False)),
        (1.0, 1.0,  (None, None), (False, True, True, False), (True, False)),
        (1.0, 1,    (None, None), (False, True, True, False), (True, False)),

        (1, 1,      (0, None), (False, True, True, False), (True, False)),
        (1, 1.0,    (0, None), (False, True, True, False), (True, False)),
        (1.0, 1.0,  (0.0, None), (False, True, True, False), (True, False)),
        (1.0, 1,    (0, None), (False, True, True, False), (True, False)),

        (1, 1,      (None, 0), (False, True, True, False), (True, False)),
        (1, 1.0,    (None, 0), (False, True, True, False), (True, False)),
        (1.0, 1.0,  (None, 0.0), (False, True, True, False), (True, False)),
        (1.0, 1,    (None, 0), (False, True, True, False), (True, False)),

        # 1+0.9/1.1---------------------------------
        (1, 0.9, (0.2, None), (True, True, True, True), (True, False)),
        (1, 0.9, (0.1, None), (True, True, True, False), (True, False)),
        (1, 0.9, (0.05, None), (True, True, False, False), (False, True)),
        (1, 0.9, (0, None), (True, True, False, False), (False, True)),
        (1, 1.1, (0, None), (False, False, True, True), (False, True)),
        (1, 1.1, (0.05, None), (False, False, True, True), (False, True)),
        (1, 1.1, (0.1, None), (False, True, True, True), (True, False)),
        (1, 1.1, (0.2, None), (True, True, True, True), (True, False)),

        # PERCENT -------
        # int+int
        (1, 0.9, (None, 20), (True, True, True, True), (True, False)),
        (1, 0.9, (None, 10), (True, True, True, False), (True, False)),
        (1, 0.9, (None, 5), (True, True, False, False), (False, True)),
        (1, 0.9, (None, 0), (True, True, False, False), (False, True)),
        (1, 1.1, (None, 0), (False, False, True, True), (False, True)),
        (1, 1.1, (None, 5), (False, False, True, True), (False, True)),
        (1, 1.1, (None, 10), (False, True, True, True), (True, False)),
        (1, 1.1, (None, 20), (True, True, True, True), (True, False)),

        # float+int
        (0.1, 0.09, (None, 20), (True, True, True, True), (True, False)),
        # (0.1, 0.09, (None, 10), (True, True, True, False), (True, False)),  # FIXME
        (0.1, 0.09, (None, 5), (True, True, False, False), (False, True)),
        (0.1, 0.09, (None, 0), (True, True, False, False), (False, True)),
        (0.1, 0.11, (None, 0), (False, False, True, True), (False, True)),
        (0.1, 0.11, (None, 5), (False, False, True, True), (False, True)),
        (0.1, 0.11, (None, 10), (False, True, True, True), (True, False)),
        (0.1, 0.11, (None, 20), (True, True, True, True), (True, False)),

        # # float+float # FIXME: finish it! resolve what to do!
        # (0.1, 0.0009, (None, 0.20), (True, True, True, True), (True, False)),
        # (0.1, 0.0009, (None, 0.10), (True, True, True, False), (True, False)),
        # (0.1, 0.0009, (None, 0.05), (True, True, False, False), (False, True)),
        # (0.1, 0.0009, (None, 0.00), (True, True, False, False), (False, True)),
        # (0.1, 0.0011, (None, 0.00), (False, False, True, True), (False, True)),
        # (0.1, 0.0011, (None, 0.05), (False, False, True, True), (False, True)),
        # (0.1, 0.0011, (None, 0.10), (False, True, True, True), (True, False)),
        # (0.1, 0.0011, (None, 0.20), (True, True, True, True), (True, False)),
    ]
)
def test__cmp_glet__single(
        source: float | int,
        other: float | int,
        accuracy_vp: tuple[float | None, float | None],
        _EXP_GLET: tuple[bool | Exception, ...],
        _EXP_EQ: tuple[bool | Exception, ...],
):
    accuracy_v, accuracy_p = accuracy_vp

    RAISED = not Victim(0)._cmp_accuracy__check_correctness(*accuracy_vp, False)

    # ACC init- meth- -----------------------------------------------
    if accuracy_v is None and accuracy_p is None:
        victim = Victim(source=source)

        Lambda(victim.cmp_gt, other).check_expected__assert(_EXP_GLET[0])
        Lambda(victim.cmp_ge, other).check_expected__assert(_EXP_GLET[1])
        Lambda(victim.cmp_le, other).check_expected__assert(_EXP_GLET[2])
        Lambda(victim.cmp_lt, other).check_expected__assert(_EXP_GLET[3])

        Lambda(victim.cmp_eq, other).check_expected__assert(_EXP_EQ[0])
        Lambda(victim.cmp_ne, other).check_expected__assert(_EXP_EQ[1])

        Lambda(operator.gt, victim, other).check_expected__assert(_EXP_GLET[0])
        Lambda(operator.ge, victim, other).check_expected__assert(_EXP_GLET[1])
        Lambda(operator.le, victim, other).check_expected__assert(_EXP_GLET[2])
        Lambda(operator.lt, victim, other).check_expected__assert(_EXP_GLET[3])

        Lambda(operator.eq, victim, other).check_expected__assert(_EXP_EQ[0])
        Lambda(operator.ne, victim, other).check_expected__assert(_EXP_EQ[1])

    # ACC init- meth+ -----------------------------------------------
    victim = Victim(source=source)

    Lambda(victim.cmp_gt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[0])
    Lambda(victim.cmp_ge, other, *accuracy_vp).check_expected__assert(_EXP_GLET[1])
    Lambda(victim.cmp_le, other, *accuracy_vp).check_expected__assert(_EXP_GLET[2])
    Lambda(victim.cmp_lt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[3])

    Lambda(victim.cmp_eq, other, *accuracy_vp).check_expected__assert(_EXP_EQ[0])
    Lambda(victim.cmp_ne, other, *accuracy_vp).check_expected__assert(_EXP_EQ[1])

    # ACC init0 meth+ -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=0)

    Lambda(victim.cmp_gt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[0])
    Lambda(victim.cmp_ge, other, *accuracy_vp).check_expected__assert(_EXP_GLET[1])
    Lambda(victim.cmp_le, other, *accuracy_vp).check_expected__assert(_EXP_GLET[2])
    Lambda(victim.cmp_lt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[3])

    Lambda(victim.cmp_eq, other, *accuracy_vp).check_expected__assert(_EXP_EQ[0])
    Lambda(victim.cmp_ne, other, *accuracy_vp).check_expected__assert(_EXP_EQ[1])


    Lambda(Victim, 1, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p).check_expected__assert(NestCmp_GLET_DigitAccuracy if not RAISED else Exception)

    if RAISED:
        return

    # ACC init+ meth- -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(victim.cmp_gt, other).check_expected__assert(_EXP_GLET[0])
    Lambda(victim.cmp_ge, other).check_expected__assert(_EXP_GLET[1])
    Lambda(victim.cmp_le, other).check_expected__assert(_EXP_GLET[2])
    Lambda(victim.cmp_lt, other).check_expected__assert(_EXP_GLET[3])

    Lambda(victim.cmp_eq, other).check_expected__assert(_EXP_EQ[0])
    Lambda(victim.cmp_ne, other).check_expected__assert(_EXP_EQ[1])

    # ACC init+ meth+ -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(victim.cmp_gt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[0])
    Lambda(victim.cmp_ge, other, *accuracy_vp).check_expected__assert(_EXP_GLET[1])
    Lambda(victim.cmp_le, other, *accuracy_vp).check_expected__assert(_EXP_GLET[2])
    Lambda(victim.cmp_lt, other, *accuracy_vp).check_expected__assert(_EXP_GLET[3])

    Lambda(victim.cmp_eq, other, *accuracy_vp).check_expected__assert(_EXP_EQ[0])
    Lambda(victim.cmp_ne, other, *accuracy_vp).check_expected__assert(_EXP_EQ[1])

    # OPERATOR -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(operator.gt, victim, other).check_expected__assert(_EXP_GLET[0])
    Lambda(operator.ge, victim, other).check_expected__assert(_EXP_GLET[1])
    Lambda(operator.le, victim, other).check_expected__assert(_EXP_GLET[2])
    Lambda(operator.lt, victim, other).check_expected__assert(_EXP_GLET[3])

    Lambda(operator.eq, victim, other).check_expected__assert(_EXP_EQ[0])
    Lambda(operator.ne, victim, other).check_expected__assert(_EXP_EQ[1])


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, other1, other2, accuracy_vp, _EXPECTED",
    argvalues=[
        # TRIVIAL ---------------------------------
        ("exc", 1, 1, (0, None), (Exception, Exception, Exception, Exception)),
        (1, "exc", 1, (0, None), (Exception, Exception, Exception, Exception)),
        (1, 1, "exc", (0, None), (False, False, Exception, Exception)),
        (1, 1, 1, ("exc", None), (Exception, Exception, Exception, Exception)),
        (1, 1, 1, (None, "exc"), (Exception, Exception, Exception, Exception)),
        (1, 1, 1, (0, 0), (Exception, Exception, Exception, Exception)),

        # 1+1--------------------------------------
        (1, 1, 1, (None, None), (False, False, False, True)),
        (1, 1.0, 1.0, (None, None), (False, False, False, True)),
        (1.0, 1.0, 1.0, (None, None), (False, False, False, True)),
        (1.0, 1, 1, (None, None), (False, False, False, True)),

        (1, 1, 1, (0, None), (False, False, False, True)),
        (1, 1.0, 1.0, (0, None), (False, False, False, True)),
        (1.0, 1.0, 1.0, (0, None), (False, False, False, True)),
        (1.0, 1, 1, (0, None), (False, False, False, True)),

        # 1+0.9/1.1---------------------------------
        (1, 0.9, 0.9, (0.2, None), (True, True, True, True)),
        (1, 0.9, 0.9, (0.1, None), (False, True, False, True)),
        (1, 0.9, 0.9, (0, None), (False, False, False, False)),
        (1, 1.1, 1.1, (0, None), (False, False, False, False)),
        (1, 1.1, 1.1, (0.1, None), (False, False, True, True)),
        (1, 1.1, 1.1, (0.2, None), (True, True, True, True)),
    ]
)
def test__cmp_glet__double(
        source: float | int,
        other1: float | int,
        other2: float | int,
        accuracy_vp: tuple[float | None, float | None],
        _EXPECTED: tuple[bool | Exception, ...],
):
    accuracy_v, accuracy_p = accuracy_vp

    RAISED = not Victim(0)._cmp_accuracy__check_correctness(*accuracy_vp, False)

    # ACC init- meth- -----------------------------------------------
    if accuracy_v is None and accuracy_p is None:
        victim = Victim(source=source)

        Lambda(victim.cmp_gtlt, other1, other2).check_expected__assert(_EXPECTED[0])
        Lambda(victim.cmp_gtle, other1, other2).check_expected__assert(_EXPECTED[1])
        Lambda(victim.cmp_gelt, other1, other2).check_expected__assert(_EXPECTED[2])
        Lambda(victim.cmp_gele, other1, other2).check_expected__assert(_EXPECTED[3])

        Lambda(gtlt, other1, victim, other2).check_expected__assert(_EXPECTED[0])
        Lambda(gtle, other1, victim, other2).check_expected__assert(_EXPECTED[1])
        Lambda(gelt, other1, victim, other2).check_expected__assert(_EXPECTED[2])
        Lambda(gele, other1, victim, other2).check_expected__assert(_EXPECTED[3])

    # ACC init- meth+ -----------------------------------------------
    victim = Victim(source=source)

    Lambda(victim.cmp_gtlt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[0])
    Lambda(victim.cmp_gtle, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[1])
    Lambda(victim.cmp_gelt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[2])
    Lambda(victim.cmp_gele, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[3])

    # ACC init0 meth+ -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=0)

    Lambda(victim.cmp_gtlt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[0])
    Lambda(victim.cmp_gtle, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[1])
    Lambda(victim.cmp_gelt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[2])
    Lambda(victim.cmp_gele, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[3])



    Lambda(Victim, 1, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p).check_expected__assert(NestCmp_GLET_DigitAccuracy if not RAISED else Exception)

    if RAISED:
        return

    # ACC init+ meth- -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(victim.cmp_gtlt, other1, other2).check_expected__assert(_EXPECTED[0])
    Lambda(victim.cmp_gtle, other1, other2).check_expected__assert(_EXPECTED[1])
    Lambda(victim.cmp_gelt, other1, other2).check_expected__assert(_EXPECTED[2])
    Lambda(victim.cmp_gele, other1, other2).check_expected__assert(_EXPECTED[3])

    # ACC init+ meth+ -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(victim.cmp_gtlt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[0])
    Lambda(victim.cmp_gtle, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[1])
    Lambda(victim.cmp_gelt, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[2])
    Lambda(victim.cmp_gele, other1, other2, *accuracy_vp).check_expected__assert(_EXPECTED[3])

    # OPERATOR -----------------------------------------------
    victim = Victim(source=source, cmp_accuracy_value=accuracy_v, cmp_accuracy_percent=accuracy_p)

    Lambda(gtlt, other1, victim, other2).check_expected__assert(_EXPECTED[0])
    Lambda(gtle, other1, victim, other2).check_expected__assert(_EXPECTED[1])
    Lambda(gelt, other1, victim, other2).check_expected__assert(_EXPECTED[2])
    Lambda(gele, other1, victim, other2).check_expected__assert(_EXPECTED[3])


# =====================================================================================================================
def test__accuracy_in_levels__only_v():
    # COULВ BE DELETED!!! not need!

    # 1=NONE
    victim = Victim(source=1, cmp_accuracy_value=None)

    assert victim >= 1
    assert victim.cmp_ge(1)
    assert victim.cmp_le(1)
    assert victim.cmp_ge(1, accuracy_value=0.1)
    assert victim.cmp_le(1, accuracy_value=0.1)

    assert not victim > 1
    assert not victim.cmp_gt(1)
    assert not victim.cmp_lt(1)
    assert victim.cmp_gt(1, accuracy_value=0.1)
    assert victim.cmp_lt(1, accuracy_value=0.1)

    # 2=0.1
    victim = Victim(source=1, cmp_accuracy_value=0.1)

    assert victim >= 1
    assert victim.cmp_ge(1)
    assert victim.cmp_le(1)
    assert victim.cmp_ge(1, accuracy_value=0.1)
    assert victim.cmp_le(1, accuracy_value=0.1)

    assert victim > 1
    assert victim.cmp_gt(1)
    assert victim.cmp_lt(1)
    assert victim.cmp_gt(1, accuracy_value=0.1)
    assert victim.cmp_lt(1, accuracy_value=0.1)

    other = 0.9
    assert victim >= other
    assert victim.cmp_ge(other)
    assert victim.cmp_le(other)
    assert victim.cmp_ge(other, accuracy_value=0.1)
    assert victim.cmp_le(other, accuracy_value=0.1)


# =====================================================================================================================



import pytest
from base_aux.buses.m1_serial3_server import *
from base_aux.buses.m1_serial1_client import *
from base_aux.base_values.m5_value_valid2_variants import *
from base_aux.base_values.m5_value_valid3_unit import *
from base_aux.aux_text.m3_parser1_cmd_args_kwargs import *


# =====================================================================================================================
@pytest.mark.xfail  # FIXME: BROKEN
class Test__SerialServer_NoConnection:
    # @classmethod
    # def setup_class(cls):
    #     pass
    #
    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    def setup_method(self, method):
        pass
        self.Victim = type("Victim", (SerialServer_Example,), {})
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__cmd__script(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("script1")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM
        assert victim._cmd__(CmdArgsKwargsParser("script script1")) == AnswerVariants.SUCCESS

    def test__LISTS(self):
        victim = self.Victim()
        assert set(victim._LIST__CMDS) == {
            "set", "get",
            "hello", "help", "echo",
            "upper", "lower",

            "cmd", "cmd_no_line",

            "script",
        }
        assert set(victim._LIST__SCRIPTS) == {
            "script1",
        }

    def test__cmd__cmd(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("cmd")) == AnswerVariants.SUCCESS
        assert victim._cmd__(CmdArgsKwargsParser("cmd_no_line")) == AnswerVariants.SUCCESS

    def test__cmd__echo(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("echo 123")) == "echo 123"
        assert victim._cmd__(CmdArgsKwargsParser("echo HELLO")) == "echo HELLO"

    def test__cmd__upper_lower(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("upper HELLO")) == "UPPER HELLO"
        assert victim._cmd__(CmdArgsKwargsParser("lower HELLO")) == "lower hello"

    def test__GET__single(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("hello123")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM

        assert victim._cmd__(CmdArgsKwargsParser("get str")) == "str"
        assert victim._cmd__(CmdArgsKwargsParser("str")) == "str"
        assert victim._cmd__(CmdArgsKwargsParser("blanc")) == ""
        assert victim._cmd__(CmdArgsKwargsParser("zero")) == '0'

        assert victim._cmd__(CmdArgsKwargsParser("int")) == '1'
        assert victim._cmd__(CmdArgsKwargsParser("float")) == '1.1'

        assert victim._cmd__(CmdArgsKwargsParser("none")) == "None"
        assert victim._cmd__(CmdArgsKwargsParser("true")) == "True"
        assert victim._cmd__(CmdArgsKwargsParser("false")) == "False"

        try:
            float(victim._cmd__(CmdArgsKwargsParser("call")))
        except:
            assert False
        assert victim._cmd__(CmdArgsKwargsParser("exc")) == AnswerVariants.ERR__PARAM_CALLING

        assert victim._cmd__(CmdArgsKwargsParser("list")) == "[0, 1, 2]"
        assert victim._cmd__(CmdArgsKwargsParser("_set")) == "{0, 1, 2}"
        assert victim._cmd__(CmdArgsKwargsParser("get _set")) == "{0, 1, 2}"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short")) == "{1: 11}"

    def test__GET__nested__list(self):
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("list")) == "[0, 1, 2]"

        assert victim._cmd__(CmdArgsKwargsParser("list 1")) == "1"
        assert victim._cmd__(CmdArgsKwargsParser("list/1")) == "1"

        assert victim._cmd__(CmdArgsKwargsParser("get list 1")) == "1"
        assert victim._cmd__(CmdArgsKwargsParser("get list/1")) == "1"

        assert victim._cmd__(CmdArgsKwargsParser("list/10")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM

        # -----
        assert victim._cmd__(CmdArgsKwargsParser("list_2")) == "[[11]]"
        assert victim._cmd__(CmdArgsKwargsParser("list_2/0")) == "[11]"
        assert victim._cmd__(CmdArgsKwargsParser("list_2/0/0")) == "11"

        assert victim._cmd__(CmdArgsKwargsParser("list_2 0 0")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("get list_2 0 0")) == "11"

    def test__GET__dict(self):
        # TODO: add BOOL/NONE
        victim = self.Victim()
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2")) == "{'HEllo': {1: 11}}"
        assert victim._cmd__(CmdArgsKwargsParser("get dict_short_2")) == "{'HEllo': {1: 11}}"

        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2/hello")) == "{1: 11}"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2 hello")) == "{1: 11}"
        assert victim._cmd__(CmdArgsKwargsParser("get dict_short_2 hello")) == "{1: 11}"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2/hello11111")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM

        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2/hello/1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("get dict_short_2/hello/1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2 hello/1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("get dict_short_2 hello/1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2 hello 1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("get dict_short_2 hello 1")) == "11"
        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2/hello/11111")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM

        assert victim._cmd__(CmdArgsKwargsParser("dict_short_2/hello/1/9/9/9/9/")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM

    @pytest.mark.xfail   # FIXME: BROKEN
    def test__SET__level_first__type(self):
        victim = self.Victim()
        assert victim.PARAMS["VAR"] == ""

        # line_parsed = CmdArgsKwargsParser("var=True")
        # result = victim._cmd__(line_parsed)
        assert victim._cmd__(CmdArgsKwargsParser("var=True")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] is True

        assert victim._cmd__(CmdArgsKwargsParser("var=false")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] is False

        assert victim._cmd__(CmdArgsKwargsParser("var=null")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] is None

        assert victim._cmd__(CmdArgsKwargsParser("var=")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == ""

        # assert victim._cmd__(CmdArgsKwargsParser("var=''")) == AnswerVariants.SUCCESS
        # assert victim.PARAMS["VAR"] == "''"                             # TODO: convert to expected!!!

        assert victim._cmd__(CmdArgsKwargsParser("var=0")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == 0

        assert victim._cmd__(CmdArgsKwargsParser("var=123")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == 123

        assert victim._cmd__(CmdArgsKwargsParser("var=1.1")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == 1.1
        assert victim._cmd__(CmdArgsKwargsParser("var=1,1")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] != 1.1

        # ITERABLES --------------------
        assert victim._cmd__(CmdArgsKwargsParser("var=[]")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == []

        # assert victim._cmd__(CmdArgsKwargsParser("var=[1, 2]")) == AnswerVariants.SUCCESS  # TODO: CLEAR ALL INTERNAL SPACES!!!
        assert victim._cmd__(CmdArgsKwargsParser("var=[1,2]")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == [1, 2]

    @pytest.mark.xfail   # FIXME: BROKEN
    def test__SET__level_first__syntax(self):
        victim = self.Victim()
        assert victim.PARAMS["INT"] == 1

        assert victim._cmd__(CmdArgsKwargsParser("set int=10")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 10

        assert victim._cmd__(CmdArgsKwargsParser("int=11")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 11

        assert victim._cmd__(CmdArgsKwargsParser("int    =12")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 12

        assert victim._cmd__(CmdArgsKwargsParser("int=     13")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 13

        assert victim._cmd__(CmdArgsKwargsParser("   int       =     14   ")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 14

        assert victim._cmd__(CmdArgsKwargsParser("   int =======     15   ")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["INT"] == 15

        # several
        assert victim._cmd__(CmdArgsKwargsParser("var 1 int=16")) == AnswerVariants.ERR__ARGS_VALIDATION
        assert victim.PARAMS["INT"] == 15

        victim.PARAMS["VAR"] = 1
        assert victim.PARAMS["VAR"] == 1
        assert victim._cmd__(CmdArgsKwargsParser("var=11 int=16")) == AnswerVariants.SUCCESS
        assert victim.PARAMS["VAR"] == 11
        assert victim.PARAMS["INT"] == 16

        # NOCHANGES if wrong name exists
        assert victim._cmd__(CmdArgsKwargsParser("var=110 int123=160")) == AnswerVariants.ERR__NAME_CMD_OR_PARAM
        assert victim.PARAMS["VAR"] == 11
        assert victim.PARAMS["INT"] == 16

    @pytest.mark.xfail   # FIXME: BROKEN
    def test__ValueUnit(self):
        victim = self.Victim()
        victim.PARAMS["UNIT123"] = ValueUnit(1, unit="V")
        assert victim.PARAMS["UNIT123"] == ValueUnit(1, unit="V")
        assert victim._cmd__(CmdArgsKwargsParser("unit123")) == "1V"

        assert victim._cmd__(CmdArgsKwargsParser("unit123=11")) == AnswerVariants.SUCCESS
        assert victim._cmd__(CmdArgsKwargsParser("unit123")) == "11V"

        assert victim._cmd__(CmdArgsKwargsParser("unit123=1.00")) == AnswerVariants.SUCCESS
        assert victim._cmd__(CmdArgsKwargsParser("unit123")) == "1.0V"

        assert victim.PARAMS["UNIT123"] == ValueUnit(1.0, unit="V")

    @pytest.mark.xfail   # FIXME: BROKEN
    def test__ValueVariants(self):
        victim = self.Victim()

        victim.PARAMS["VARIANT"] = ValueVariants(220, variants=[220, 380])
        assert victim._cmd__(CmdArgsKwargsParser("variant")) == "220"

        assert victim._cmd__(CmdArgsKwargsParser("variant=11")) == AnswerVariants.ERR__VALUE_INCOMPATIBLE
        assert victim._cmd__(CmdArgsKwargsParser("variant")) == "220"

        assert victim._cmd__(CmdArgsKwargsParser("variant=380")) == AnswerVariants.SUCCESS
        assert victim._cmd__(CmdArgsKwargsParser("variant")) == "380"

        assert victim.PARAMS["VARIANT"] == ValueVariants(380, variants=[220, 380])

    def test__list_results(self):
        victim = self.Victim()
        victim.PARAMS["VARIANT"] = ValueVariants(220, variants=[220, 380])
        victim.PARAMS["UNIT123"] = ValueUnit(1, unit="V")

        assert victim.list_param_results(["VARIANT", "UNIT123"]) == ["VARIANT=220", "UNIT123=1V"]
        assert victim.list_param_results(["cmd", "unit123"]) == [f"cmd={AnswerVariants.SUCCESS}", "unit123=1V"]


# =====================================================================================================================
@pytest.mark.xfail  # FIXME: BROKEN
class Test_SerialServer_WithConnection:
    Victim: type[SerialClient] = type("Victim", (SerialClient,), {})
    victim: SerialClient = None

    @classmethod
    def setup_class(cls):
        if cls.Victim.addresses_paired__count() < 1:
            msg = f"[ERROR] need connect TWO SerialPorts"
            print(msg)
            raise Exception(msg)

    @classmethod
    def teardown_class(cls):
        pass
        if cls.victim:
            cls.victim.disconnect()

    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__1(self):
        # server start
        server = SerialServer_Example()
        server.connect()

        class Victim(SerialClient):
            _ADDRESS = Enum__AddressAutoAcceptVariant.FIRST_FREE__ANSWER_VALID

            def address__validate(self) -> bool | None | NoReturn:
                return server.HELLO_MSG[0] in self.write_read("hello").list_output()

        victim = Victim()
        assert victim.connect()

        assert victim.write_read__last("echo 123") == "echo 123"
        assert victim.write_read__last("CMD_NOT_ESISTS") == server.ANSWER.ERR__NAME_CMD_OR_PARAM
        assert victim.write_read__last("upper hello") == "UPPER HELLO"


# =====================================================================================================================



from typing import *

from base_aux.tables.m1_table_obj import TableObj


# =====================================================================================================================
class TableColumn:
    """
    GOAL
    ----
    get item by name+index_column from Table
    just an attempt for simplification access to exact index item from list

    now you can access objects directly by
        dev_table_column.ATC
    instead old
        dev_table.ATC[index]
    """
    table: TableObj | dict[Any, list | Collection]
    column: int

    def __init__(self, *, column: int, table: TableObj | dict[Any, list | Collection]) -> None:
        self.table = table
        self.column = column

    def __getitem__(self, item: str) -> Any | NoReturn:
        return self.table[item][self.column]

    def __getattr__(self, item: str) -> Any | NoReturn:
        return self.table[item][self.column]


# =====================================================================================================================



from typing import *

import pathlib
import json
import functools

import requests
from aiohttp import web
import aiohttp
import yaml

from PyQt5.QtCore import QThread


# =====================================================================================================================
TYPE__SELF = Any
TYPE__REQUEST = Any


# =====================================================================================================================
class Exc__AiohttpServerStartSameAddress(Exception):
    pass


class Exc__LinuxPermition(Exception):
    pass


class Exc__AiohttpServerOtherError(Exception):
    pass


# =====================================================================================================================
def decorator__log_request_response(func: Callable[[TYPE__SELF, TYPE__REQUEST], Coroutine[Any, Any, web.Response]]):
    @functools.wraps(func)
    async def _wrapper(self, request):
        # ObjectInfo(request).print()

        response = await func(self, request)
        # ObjectInfo(response).print()
        # print(f"API access[from={request.remote=}/to={request.host=}][{self.__class__.__name__}.{func.__name__}()][{response.status=}]")
        return response

    return _wrapper


# =====================================================================================================================
class ServerAiohttpBase(QThread):
    """
    NOTE: some problems withing PYTEST only! you can do any instances in one program!

    SEE testplans.TpApi as example
    """
    # SETTINGS -----------------------------
    START_ON_INIT: bool = False
    CONFIG_FILEPATH: Union[pathlib.Path, str] = pathlib.Path(__file__).parent / 'server_aiohttp__config.yaml'
    PORT: Optional[int] = 80  # None==8080/directWeb==80

    # AUX ----------------------------------
    _ROUTE_FUNC_START_PATTERN: str = "response_%s__"
    _ROUTE_NAME_PREFIX_HTML_FOR_JSON: str = "html__"
    _app: web.Application
    _ROUTES: dict[str, list[str]] = {
        "get_html": [],
        "get_json": [],
        "post": [],
    }
    data: Any

    def __init__(self, data: Any = None):
        """
        :param data: just to keep as further standard
        """
        super().__init__()
        self.CONFIG_FILEPATH: pathlib.Path = pathlib.Path(self.CONFIG_FILEPATH)
        self.data = data

        self._app: web.Application = web.Application()
        self.setup_routes()
        self.apply_config()

    # =================================================================================================================
    def run(self) -> None:
        """
        NOTE: this will block process!
        but if start() in thread - it would be OK!

        EXC will not catch from start!!! but will CAUSE SYS_EXIT!!!
        """
        try:
            web.run_app(app=self._app, port=self.PORT)
            # NOTE:
            # 1. dont use parameter host="localhost" - its incorrect! from other host you cant access by IP!!! - if not specified - OK!

            # this will not catch!!! cause of thread maybe!!!
        except PermissionError as exc:
            # PermissionError(13, "error while attempting to bind on address ('127.0.0.1', 80): permission denied"
            msg = f"[ERROR] need linux rights for accessing ports under 1024 (execute [sudo setcap CAP_NET_BIND_SERVICE=+eip /usr/bin/python3.11]) or use appropriate port {exc!r}"
            raise Exc__LinuxPermition(msg)

        except OSError as exc:
            # OSError(10048, "error while attempting to bind on address ('0.0.0.0', 80)
            msg = f"[ERROR]started same server address/port {exc!r}"
            raise Exc__AiohttpServerStartSameAddress(msg)

        except Exception as exc:
            msg = f"[ERROR] other error {exc!r}"
            raise Exc__AiohttpServerOtherError(msg)  # DON'T DELETE RAISE! - IT IS VERY NECESSARY/IMPORTANT for tests!

    def start(self, *args):
        if not self.isRunning():
            super().start()

    def ___server_pause(self):
        """
        just as info
        """
        self._app.shutdown()    # pause server!
        self._app.startup()

    # =================================================================================================================
    def apply_config(self, config_filepath=None):
        config_filepath = config_filepath or self.CONFIG_FILEPATH
        if config_filepath and config_filepath.exists():
            with open(config_filepath) as f:
                config = yaml.safe_load(f)
                self._app["config"] = config
            print(f"{self._app['config']=}")
        else:
            msg = f"[WARN] no file {self.CONFIG_FILEPATH=}"
            print(msg)
            self._app["config"] = {}

        # self._app['config']={'postgres': {'user': 'aiohttpdemo_user', 'password': 'aiohttpdemo_pass', 'host': 'localhost', 'port': 5432}}

    # =================================================================================================================
    def setup_routes(self):
        for attr_name in dir(self):
            for group_name, group_list in self._ROUTES.items():
                group_name_start = self._ROUTE_FUNC_START_PATTERN % group_name
                if not attr_name.startswith(group_name_start):
                    continue

                route_name_wo_slash = f'{attr_name.replace(group_name_start, "")}'
                route_name_w_slash = f'/{route_name_wo_slash}'
                self._ROUTES[group_name].append(route_name_w_slash)
                if group_name == 'get_html':
                    self._app.router.add_get(route_name_w_slash, getattr(self, attr_name))
                elif group_name == 'get_json':
                    self._app.router.add_get(route_name_w_slash, getattr(self, attr_name))
                    self._app.router.add_get(f"/{self._ROUTE_NAME_PREFIX_HTML_FOR_JSON}{route_name_wo_slash}", self._response_get_json__converted_to_html)

                    self._ROUTES["get_html"].append(f"/{self._ROUTE_NAME_PREFIX_HTML_FOR_JSON}{route_name_wo_slash}")

                elif group_name == 'post':
                    self._app.router.add_post(route_name_w_slash, getattr(self, attr_name))
                    self._app.router.add_get(route_name_w_slash, self._response_post__converted_to_get)

    # =================================================================================================================
    def html_create(
            self,
            data: Union[None, str, dict] = None,
            redirect_time: Optional[int] = None,
            redirect_source: Optional[str] = None,
            request: Any = None,
    ) -> str:
        route = request.path
        data = data or ""

        # -------------------------------------
        part_refresh = ""
        if redirect_time:
            part_refresh = f"<meta http-equiv='refresh' content='{redirect_time}; url=/' />"

        # -------------------------------------
        if isinstance(data, str):
            pass

        elif isinstance(data, dict):
            data = json.dumps(data, indent=4)
            # body_html = body_pretty.replace("\n", "<br />\n")     # dont need!
            data = f"<pre>{data}</pre>"

        else:
            data = str(data)

        # -------------------------------------
        body_header = f"<a href='/'>HOME</a>{route}<br />"
        if request:
            body_header += f"request from={request.remote=}/to={request.host=}<br />"

        # -------------------------------------
        result = f"""
        <!doctype html>
        <html lang="en-US">
            <head>
                <meta charset="utf-8" />
                <title>{route}</title>
                {part_refresh}
            </head>
            <body>
                <p>{body_header}</p>
                <p>{data}</p>
            </body>
        </html>
        """
        return result

    def html_block__api_index(self) -> str:
        html_block = f""
        for group, names in self._ROUTES.items():
            html_block += f"{group.upper()}:<br />"
            for name in names:
                html_block += f"<a href='{name}'>{name}</a><br />"

            html_block += f"<br />"
        return html_block

    # =================================================================================================================
    async def _response_post__converted_to_get(self, request) -> web.Response:
        route = request.path[1:]
        await getattr(self, self._ROUTE_FUNC_START_PATTERN % "post" + route)(request)

        # RESPONSE --------------------------------------------------
        html = self.html_create(data="", request=request, redirect_time=1)
        return web.Response(text=html, content_type='text/html')

    async def _response_get_json__converted_to_html(self, request) -> web.Response:
        route_json = request.path[1:].replace(self._ROUTE_NAME_PREFIX_HTML_FOR_JSON, "")
        response: requests.Response = await getattr(self, self._ROUTE_FUNC_START_PATTERN % "get_json" + route_json)(request)
        data_text = response.text   # no json/data here in instance!
        data_json = json.loads(data_text)

        # RESPONSE --------------------------------------------------
        html = self.html_create(data=data_json, request=request)
        return web.Response(text=html, content_type='text/html')

    # =================================================================================================================
    async def response_get_html__(self, request) -> web.Response:
        return await self.response_get_html__api_index(request)

    async def response_get_html__api_index(self, request) -> web.Response:
        # RESPONSE --------------------------------------------------
        html = self.html_create(data=self.html_block__api_index(), request=request)
        return web.Response(text=html, content_type='text/html')

    # THIS IS AS HELP COMMENT!
    # @decorator__log_request_response
    # async def response_post__start(self, request) -> web.Response:
    #     # return self.response_get__start(request)  # this is will not work!
    #     self.data.signal__tp_start.emit()
    #     test_data = await request.json()  # dont use

    #     # RESPONSE --------------------------------------------------
    #     response = web.json_response(data={})
    #     return response


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda import *
from base_aux.aux_text.m1_text_aux import TextAux
from base_aux.base_enums.m2_enum1_adj import *
from base_aux.aux_text.m0_text_examples import *


# =====================================================================================================================
class Test__sub:
    @pytest.mark.parametrize(
        argnames="source, rule, _EXPECTED",
        argvalues=[
            # NOT ACCEPTED -------------
            # ("None123", ("None00"), "None123"),
            # ("None123", ("None"), "123"),
            # ("None123", ("None", None), "123"),
            # ("None123", ("None", "0"), "0123"),
            #
            # ("1*3", ("\*", "2"), "123"),
            ("1:1", (r"\b(\d+\.?\d*)\b\s*:\s*", r'"\1":'), '"1":1'),
            ("123:1", (r"\b(\d+\.?\d*)\b\s*:\s*", r'"\1":'), '"123":1'),
        ]
    )
    def test__regexp(self, source, rule, _EXPECTED):
        func_link = TextAux(source).sub__regexp
        Lambda(func_link, *rule).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, rule, _EXPECTED",
        argvalues=[
            # NOT ACCEPTED -------------
            ("None123", ("None", "null"), "None123"),
            ("None_123", ("None", "null"), "None_123"),

            # ACCEPTED -------------
            ("null", ("None", "null"), "null"),
            ("None", ("None", "null"), "null"),
            ("None-123", ("None", "null"), "null-123"),

            # CONTAINERS -------------
            ("[null]", ("None", "null"), "[null]"),
            ("[None]", ("None", "null"), "[null]"),
            ("[None, ]", ("None", "null"), "[null, ]"),

            (" None, 123", ("None", "null"), " null, 123"),
            ("[None, null, 123]", ("None", "null"), "[null, null, 123]"),

            ("[None, null, 323]", (r"None.*3", "null"), "[null]"),

            ("[None, null, 3 23]", (r"None.*3", "null"), "[null]"),
            ("[None, null, 3 23]", (r"None.*?3", "null"), "[null 23]"),
        ]
    )
    def test__words(self, source, rule, _EXPECTED):
        func_link = TextAux(source).sub__word
        Lambda(func_link, *rule).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__Edit:
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (" " * 2, ""),
            (" " * 3, ""),
            ("line   line", "lineline"),

            (" \n ", "\n"),

            (" 1, _\n\n  \n \t  \r  ()", '1,_\n\n\n\t\r()'),
        ]
    )
    def test__spaces_all(self, source, _EXPECTED):
        func_link = TextAux(source).clear__spaces_all
        Lambda(func_link).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (" " * 2, " "),
            (" " * 3, " "),
            ("line   line", "line line"),

            (" \n  ", " \n "),

            (" 1, _\n\n  \n \t  \r  ()", ' 1, _\n\n \n \t \r ()'),
        ]
    )
    def test__spaces_duplicates(self, source, _EXPECTED):
        func_link = TextAux(source).clear__space_duplicates
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, pat, _EXPECTED",
        argvalues=[
            (" line1 \n \n line3 ", r".*1.*", "\n \n line3 "),
            (" line1 \n \n line3 ", r".*3.*", " line1 \n \n"),
            # (" line1 \n \t\r\n line3 ", r"", " line1 \n line3 "),
        ]
    )
    def test__lines(self, source, pat, _EXPECTED):
        func_link = TextAux(source).clear__lines
        Lambda(func_link, pat).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            # BLANK -----
            ("\n", ""),
            (" \n ", ""),

            # EDGES -----
            ("\n line1", " line1"),
            (" \n line1", " line1"),
            (" line1 \n ", " line1 "),

            # MIDDLE -----
            (" line1 \n\n \nline2 ", " line1 \nline2 "),

            # OTHER -----
            (" line1 \n \n line3 ", " line1 \n line3 "),
            (" line1 \n \t\r\n line3 ", " line1 \n line3 "),
        ]
    )
    def test__lines_blank(self, source, _EXPECTED):
        func_link = TextAux(source).delete__lines_blank
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, cmt, _EXPECTED",
        argvalues=[
            # EnumAdj_CmtStyle.SHARP -------------------------------------------------------
            # ZERO -----
            ("line1 ", EnumAdj_CmtStyle.SHARP, "line1 "),

            # SEPARATED -----
            ("#cmt #", EnumAdj_CmtStyle.SHARP, ""),
            ("##cmt #", EnumAdj_CmtStyle.SHARP, ""),

            ("#cmt ", EnumAdj_CmtStyle.SHARP, ""),
            ("  # cmt 1 ", EnumAdj_CmtStyle.SHARP, ""),

            # INLINE -----
            ("line  # cmt 1 ", EnumAdj_CmtStyle.SHARP, "line"),

            # SEVERAL LINES ====
            ("line1  # cmt1 \n line2 ", EnumAdj_CmtStyle.SHARP, "line1\n line2 "),
            ("line1  # cmt1 \n line2 #cmt2", EnumAdj_CmtStyle.SHARP, "line1\n line2"),
            ("line1  # cmt1 \n #cmt \n line2 #cmt2", EnumAdj_CmtStyle.SHARP, "line1\n line2"),

            # EnumAdj_CmtStyle.REM ---------------------------------------------------------
            # ZERO -----
            ("line1 ", EnumAdj_CmtStyle.REM, "line1 "),

            # SEPARATED -----
            ("REM #", EnumAdj_CmtStyle.REM, ""),
            ("REM  REM #", EnumAdj_CmtStyle.REM, ""),

            # INLINE -----
            ("line  REM 1 ", EnumAdj_CmtStyle.REM, "line"),

            # SEVERAL LINES ====
            ("line1  REM cmt1 \n line2 ", EnumAdj_CmtStyle.REM, "line1\n line2 "),
            ("line1  REM cmt1 \n line2 REM", EnumAdj_CmtStyle.REM, "line1\n line2 REM"),
            ("line1  REM cmt1 \n line2 REM ", EnumAdj_CmtStyle.REM, "line1\n line2"),
            ("line1  REM cmt1 \n REM \n line2 REM ", EnumAdj_CmtStyle.REM, "line1\n line2"),

            # EnumAdj_CmtStyle.C ---------------------------------------------------------
            ("line1  /*cmt*/  \n /*cmt*/  \n line2 /*cmt*/  ", EnumAdj_CmtStyle.C, "line1\n line2"),
            ("line1  /*cmt*/  \n /*\ncmt*/  \n line2 /*cmt*/  ", EnumAdj_CmtStyle.C, "line1\n line2"),
        ]
    )
    def test__clear__cmts(self, source, cmt, _EXPECTED):
        func_link = TextAux(source).delete__cmts
        Lambda(func_link, cmt).check_expected__assert(_EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            ("  ", ("", "", "")),
            ("line1  ", ("line1", "line1  ", "line1")),
            ("  line1 ", ("line1", "line1 ", "  line1")),

            ("line1  \n  line2", ("line1\nline2", "line1  \nline2", "line1\n  line2")),
            ("line1  \n  \n  line2", ("line1\nline2", "line1  \nline2", "line1\n  line2")),
        ]
    )
    def test__strip__lines(self, source, _EXPECTED):
        Lambda(TextAux(source).strip__lines).check_expected__assert(_EXPECTED[0])
        Lambda(TextAux(source).lstrip__lines).check_expected__assert(_EXPECTED[1])
        Lambda(TextAux(source).rstrip__lines).check_expected__assert(_EXPECTED[2])


# =====================================================================================================================
class Test__find:
    @pytest.mark.parametrize(
        argnames="source, patt, _EXPECTED",
        argvalues=[
            ("None123", r"\w*", ["None123", ]),
            ("None123", r"\w+(?#*.*)", ["None123", ]),
            ("None123  #cmt", r"\w+", ["None123", "cmt"]),
            ("   None123  #cmt", r"\w+", ["None123", "cmt"]),
        ]
    )
    def test__1(self, source, patt, _EXPECTED):
        func_link = TextAux(source).findall
        Lambda(func_link, patt).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__shortcut:
    @pytest.mark.parametrize(
        argnames="p1,p2,p3,p4,_EXPECTED",
        argvalues=[
            (None, 5, "...", EnumAdj_Where3.LAST, "None"),  # DOES NOT METTER!
            ("", 5, "...", EnumAdj_Where3.LAST, ""),
            ("123456", 3, "...", EnumAdj_Where3.LAST, "..."),

            ("123", 3, "#", EnumAdj_Where3.LAST, "123"),

            ("1223", 3, "#", EnumAdj_Where3.FIRST, "#23"),
            ("1223", 3, "#", EnumAdj_Where3.MIDDLE, "1#3"),
            ("1223", 3, "#", EnumAdj_Where3.LAST, "12#"),

            ("1223", 3, "##", EnumAdj_Where3.FIRST, "##3"),
            ("1223", 3, "##", EnumAdj_Where3.MIDDLE, "##3"),
            ("1223", 3, "##", EnumAdj_Where3.LAST, "1##"),

            ("1223", 3, "###", EnumAdj_Where3.FIRST, "###"),
            ("1223", 3, "###", EnumAdj_Where3.MIDDLE, "###"),
            ("1223", 3, "###", EnumAdj_Where3.LAST, "###"),

            ("1223", 1, "###", EnumAdj_Where3.FIRST, "#"),
            ("1223", 1, "###", EnumAdj_Where3.MIDDLE, "#"),
            ("1223", 1, "###", EnumAdj_Where3.LAST, "#"),

            #
            ("123", 1, "##", EnumAdj_Where3.FIRST, "#"),
            ("123", 2, "##", EnumAdj_Where3.FIRST, "##"),
            ("123", 3, "##", EnumAdj_Where3.FIRST, "123"),
            ("123", 4, "##", EnumAdj_Where3.FIRST, "123"),
            ("123", 5, "##", EnumAdj_Where3.FIRST, "123"),
        ]
    )
    def test__1(self, p1,p2,p3,p4,_EXPECTED):
        func_link = TextAux(p1).shortcut
        Lambda(func_link, **dict(maxlen=p2, sub=p3, where=p4)).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__try_convert_to_object:
    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (0, True),
            (1, True),
            (10, True),
            (1.0, True),
            (1.1, True),
            (-1.1, True),
            ("", False),
            ('', False),
            ("hello", False),
            (None, True),
            (True, True),
            (False, True),
            ([None, True, 1, -1.1, "hello", "", ''], True),
            # [None, True, 1, -1.1, "hello", [], {1:1}],  #JSONDecodeError('Expecting property name enclosed in double quotes: line 1 column 37 (char 36)')=KEYS IS ONLY STRINGS
            ([None, True, 1, -1.1, "hello", [], {"1": 1, "hello": []}], True),

            ([], True),
            # (),     # JSONDecodeError('Expecting value: line 1 column 1 (char 0)') - НЕСУЩЕСТВУЕТ TUPLE???
            ({}, True),
            ({"key": True}, True),
            ({"key": False}, True),
            ({"key": None}, True),
            ({"key": 123}, True),
        ]
    )
    def test__MAIN_GOAL__string_source(self, source, _EXPECTED):
        func_link = TextAux(str(source)).parse__object_stringed  # DONT DELETE STR!!!

        assert Lambda(func_link).check_expected__bool(source) == _EXPECTED      # IT IS CORRECT! dont change!
        # Lambda(func_link, source).check_expected__assert(_EXPECTED)           # it is not the same!!!

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            # SINGLE ---------------------
            ({"key1": 123}, {"key1": 123}),

            ({"1": 123}, {"1": 123}),
            ({1: 123}, {"1": 123}),

            ({"1.1": 123}, {"1.1": 123}),
            ({1.1: 123}, {"1.1": 123}),

            # SEVERAL ---------------------
            ({1:1,2:2,3.3:3}, {"1": 1, "2": 2, "3.3": 3}),
            ({1:{1:1},2:{2:2},3.3:3}, {"1":{"1":1}, "2":{"2":2}, "3.3": 3}),
        ]
    )
    def test__parse_dict_keys(self, source, _EXPECTED):
        func_link = TextAux(source).parse__object_stringed
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    def base_test__try_convert_to_object(self, source, _EXPECTED):
        func_link = TextAux(source).parse__object_stringed
        Lambda(func_link).check_expected__assert(_EXPECTED)

    # =================================================================================================================
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, None),
            (True, True),
            (False, False),
            (0, 0),
            ([], []),
            ({"1": 1}, {"1": 1}),   # FIXME: need ref to use always ints in keys!
            # ({1: 1}, {1: 1}),   # FIXME: need ref to use always ints in keys!
        ]
    )
    def test__originals(self, source, _EXPECTED):
        self.base_test__try_convert_to_object(source, _EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            ("null", None),
            ("true", True),
            ("false", False),

            ("None", None),

            # ("['None', 1]", [None, 1]),  #JSONDecodeError('Expecting value: line 1 column 2 (char 1)')    # VALUES only DOUBLEQUOTS!!! accepted!!!    # FIXME:
            # ('["None", 1]', [None, 1]),  #FIXME: dont replace None in any quats
            ("[null, 1]", [None, 1]),
            ("[None, 1]", [None, 1]),

            ("True", True),
            ("False", False),
        ]
    )
    def test__bools(self, source, _EXPECTED):
        self.base_test__try_convert_to_object(source, _EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (0, 0),
            ("0", 0),
            ("00", NoValue),
            ("01", NoValue),
            ("10", 10),

            ("1.0", 1.0),
            ("1.000", 1.0),

            ("1,000", NoValue),
        ]
    )
    def test__nums(self, source, _EXPECTED):
        self.base_test__try_convert_to_object(source, _EXPECTED)

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            ("[]", []),
            # ("()", ()),     #JSONDecodeError('Expecting value: line 1 column 1 (char 0)'
            ("{}", {}),
            # ("{1: 1}", {1: 1}), #JSONDecodeError('Expecting property name enclosed in double quotes: line 1 column 2 (char 1)'
            # ("{'1': 1}", {'1': 1}),   #JSONDecodeError('Expecting property name enclosed in double quotes: line 1 column 2 (char 1)'
            ('{"1": 1}', {'1': 1}),  # KEYS ONLY in DOUBLEQUOTS!!!
        ]
    )
    def test__containers(self, source, _EXPECTED):
        self.base_test__try_convert_to_object(source, _EXPECTED)


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        ("None1", ["None1", ]),
        ("None1   #cmt", ["None1", ]),
        ("   None1   #cmt", ["None1", ]),
        ("#cmt      None2", []),
        ("#cmt \n \n None1 #cmt   hello \n   None2", ["None1", "None2", ]),

        ("None>=1.1", ["None>=1.1", ]),
        ("None>=1.1 #cmt \n None=[1.2, ]", ["None>=1.1", "None=[1.2, ]"]),
    ]
)
def test__requirements__get_list(source, _EXPECTED):
    func_link = TextAux(source).parse__requirements_lines
    Lambda(func_link).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__ParseNum:
    @pytest.mark.parametrize(
        argnames="source, fpoint, _EXPECTED",
        argvalues=[
            (0, None, 0),
            ("0", None, 0),
            (123, None, 123),
            ("123", None, 123),
            (1.1, None, 1.1),
            ("1.1", None, 1.1),
            ("1,1", None, 1.1),
            ("1,1.", None, 1.1),
            ("1.1.", None, 1.1),
            ("1.1,", None, 1.1),

            ("   0   ", None, 0),
            ("   000   ", None, 0),
            ("-aa001cc", None, 1),
            ("-aa001.000cc", None, 1.0),
            ("-aa001,000cc", None, 1.0),
            ("   -aa001.200cc   ", None, 1.2),
            ("   -aa001.200cc   ", ",", None),
            ("   -aa001,200cc   ", ",", 1.2),
            ("   -aa--001,200cc   ", ",", -1.2),
        ]
    )
    def test__num(self, source, fpoint, _EXPECTED):
        func_link = TextAux(source).parse__number_single
        Lambda(func_link, fpoint).check_expected__assert(_EXPECTED)


# =====================================================================================================================
class Test__ParseDict:
    @pytest.mark.parametrize(
        argnames="source, _EXPECTED",
        argvalues=[
            (None, None),
            (0, None),
            (1, None),
            ("", None),
            ("1", None),
            (INI_EXAMPLES.INT_KEY__TEXT, INI_EXAMPLES.INT_KEY__DICT_MERGED),
            (INI_EXAMPLES.MESHED__TEXT, INI_EXAMPLES.MESHED__DICT_MERGED),
            (INI_EXAMPLES.NOT_MESHED__TEXT, INI_EXAMPLES.NOT_MESHED__DICT_MERGED),
        ]
    )
    def test__ini(self, source, _EXPECTED):
        Lambda(TextAux(source).parse__dict_ini).check_expected__assert(_EXPECTED)
        Lambda(TextAux(source).parse__dict).check_expected__assert(_EXPECTED)


# =====================================================================================================================



from base_aux.base_enums.m1_enum0_nest_eq import *


# =====================================================================================================================
class EnumAdj_SingleMultiple(NestEq_EnumAdj):
    NOT_EXISTS = None
    SINGLE = 1
    MULTIPLE = 2


class EnumAdj_StaticCallable(NestEq_EnumAdj):
    STATIC = 1
    CALLABLE = 2


# =====================================================================================================================
class EnumAdj_When2(NestEq_EnumAdj):
    BEFORE = 1
    AFTER = 2


class EnumAdj_When3(NestEq_EnumAdj):
    BEFORE = 1
    AFTER = 2
    MIDDLE = 3


# ---------------------------------------------------------------------------------------------------------------------
class EnumAdj_Where2(NestEq_EnumAdj):
    FIRST = 1
    LAST = 2


class EnumAdj_Where3(NestEq_EnumAdj):
    FIRST = 1
    LAST = 2
    MIDDLE = 3


# =====================================================================================================================
class EnumAdj_CallResolveStyle(NestEq_EnumAdj):
    DIRECT = 1
    EXC = 2
    RAISE = 3
    RAISE_AS_NONE = 4
    BOOL = 5

    SKIP_CALLABLE = 6
    SKIP_RAISED = 7


class EnumAdj_ReturnOnRaise(NestEq_EnumAdj):
    """
    GOAL
    ----
    select what to return on raise

    SPECIALLY CREATED FOR
    ---------------------
    ChainResolve
    """
    RAISE = 1
    NONE = 2
    EXC = 3


# =====================================================================================================================
class EnumAdj_SourceOrigOrCopy(NestEq_EnumAdj):
    """
    GOAL
    ----
    define where work process in original source or copy

    SPECIALLY CREATED FOR
    ---------------------
    DictAuxInline/Deepcopy
    """
    ORIGINAL = True
    COPY = False


# =====================================================================================================================
class EnumAdj_IgnoreCase(NestEq_EnumAdj):
    """
    GOAL
    ----
    replace many bool params about case sensing

    SPECIALLY CREATED FOR
    ---------------------
    Meta_ClsGaAnnotNamesAsValuesIc
    """
    IGNORECASE = True
    CASESENSE = False


class EnumAdj__TextCaseStyle(NestEq_EnumAdj):
    """
    GOAL
    ----
    select the representation for text

    SPECIALLY CREATED FOR
    ---------------------
    StrIc
    """
    ORIGINAL: int = 0
    LOWER: int = 1
    UPPER: int = 2


# =====================================================================================================================
class EnumAdj_ProcessResult(NestEq_EnumAdj):
    """
    GOAL
    ----
    define special values for methods

    SPECIALLY CREATED FOR
    ---------------------
    CallableAux.resolve when returns SKIPPED like object!
    """
    NONE = None
    SKIPPED = 1
    STOPPED = 2
    RAISED = 3
    FAILED = False
    SUCCESS = True


class EnumAdj_ProcessStateActive(NestEq_EnumAdj):
    """
    NAME
    ----
    STATE_ACTIVE
    """
    NONE = None
    STARTED = True
    FINISHED = False


class EnumAdj_ProcessStateResult(NestEq_EnumAdj):
    """
    GOAL
    ----
    use processActive+Result in one value

    SPECIALLY CREATED FOR
    ---------------------
    1/ VALID
    2/ tc.startup_cls/teardown_cls
    """
    NONE = None
    STARTED = 0
    FINISHED_FAIL = False
    FINISHED_SUCCESS = True


# =====================================================================================================================
class EnumAdj_StopResetHardSoft(NestEq_EnumAdj):
    SOFT: int = 1
    HARD: int = 2


# =====================================================================================================================
class EnumAdj_BoolCumulate(NestEq_EnumAdj):
    """
    GOAL
    ----
    combine result for collection

    SPECIALLY CREATED FOR
    ---------------------
    EqValid_RegexpAllTrue
    """
    ALL_TRUE = all
    ANY_TRUE = any
    ALL_FALSE = 1
    ANY_FALSE = 2


class EnumAdj_PathType(NestEq_EnumAdj):
    FILE = 1
    DIR = 2
    ALL = 3


# class AppendType(NestEq_EnumAdj):
#     NEWLINE = 1


class EnumAdj_AttemptsUsage(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    Base_ReAttempts/RExp
    """
    FIRST = None
    ALL = all


# =====================================================================================================================
class EnumAdj_DictTextFormat(NestEq_EnumAdj):
    AUTO = None     # by trying all variants
    EXTENTION = 0

    CSV = "csv"
    INI = "ini"
    JSON = "json"
    STR = "str"     # str(dict)


class EnumAdj_TextStyle(NestEq_EnumAdj):
    ANY = any       # keep decide?
    AUTO = None     # keep decide?

    CSV = "csv"
    INI = "ini"
    JSON = "json"
    TXT = "txt"

    PY = "py"
    C = "c"
    BAT = "bat"
    SH = "sh"

    REQ = "requirements"
    GITIGNORE = "gitignore"
    MD = "md"


class EnumAdj_CmtStyle(NestEq_EnumAdj):
    """
    GOAL
    ----
    select
    """
    AUTO = None     # keep decide?
    ALL = all

    SHARP = "#"
    DSLASH = "//"
    REM = "rem"
    C = "c"             # /*...*/
    SEMICOLON = ";"     # for INI files


class EnumAdj_PatCoverStyle(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    TextAux.sub__regexp
    """
    NONE = None
    WORD = "word"
    LINE = "line"


# =====================================================================================================================
class EnumAdj_NumType(NestEq_EnumAdj):
    INT = int
    FLOAT = float
    BOTH = None


class EnumAdj_NumFPoint(NestEq_EnumAdj):
    """
    GOAL
    ----
    floating point style

    SPECIALLY CREATED FOR
    ---------------------
    TextAux.parse__single_number
    """
    DOT = "."
    COMMA = ","
    AUTO = None     # auto is more important for SingleNum!


TYPING__FPOINT_DRAFT = EnumAdj_NumFPoint | str | None


# =====================================================================================================================
class EnumAdj_CmpType(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    path1_dirs.DirAux.iter(timestamp)
    """
    LT = 1
    LE = 2
    GT = 3
    GE = 4


# =====================================================================================================================
class EnumAdj_FormIntExt(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    AttrAux_Existed show internal external names for PRIVATES
    """
    INTERNAL = 1
    EXTERNAL = 2


class EnumAdj_AttrAnnotsOrExisted(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    NestInit_AnnotsAttrsByKwArgs_Base for separating work with - TODO: DEPRECATE?
    """
    ATTRS_EXISTED = None
    ANNOTS_ONLY = 1


class EnumAdj_AnnotsDepthAllOrLast(NestEq_EnumAdj):
    """
    GOAL
    ----
    need to separate work with last/allNested annots!

    SPECIALLY CREATED FOR
    ---------------------
    Base_ReqCheckStr
    """
    ALL_NESTED = None
    LAST_CHILD = 1


class EnumAdj_AttrScope(NestEq_EnumAdj):
    """
    SPECIALLY CREATED FOR
    ---------------------
    AttrKit_Blank
    """
    NOT_HIDDEN = None
    NOT_PRIVATE = 1
    ALL = 2

    PRIVATE = 3    # usually not used! just in case!


# =====================================================================================================================
# class Represent(NestEq_EnumNestEqIc_Enum):
#     NAME = 1
#     OBJECT = 2


# =====================================================================================================================
def _examples() -> None:
    WHEN = EnumAdj_When2.BEFORE
    if WHEN is EnumAdj_When2.BEFORE:
        pass

    print(EnumAdj_NumFPoint.COMMA.name)
    print(EnumAdj_NumFPoint.COMMA.value)
    print()
    print()

    print(EnumAdj_NumFPoint.COMMA)  # EnumAdj_NumFPoint.COMMA
    print(EnumAdj_NumFPoint("."))  # EnumAdj_NumFPoint.DOT

    print("." in EnumAdj_NumFPoint)  # True
    print(EnumAdj_NumFPoint.DOT in EnumAdj_NumFPoint)  # True

    print(EnumAdj_NumFPoint(".") == ".")  # True
    print(EnumAdj_NumFPoint(EnumAdj_NumFPoint.DOT))  # EnumAdj_NumFPoint.DOT     # BEST WAY to init value!


# =====================================================================================================================
if __name__ == "__main__":
    _examples()


# =====================================================================================================================



import pytest

from base_aux.aux_text.m3_parser1_cmd_args_kwargs import *


# =====================================================================================================================
class Test__LineParsed:
    @classmethod
    def setup_class(cls):
        pass
        cls.Victim = type("Victim", (CmdArgsKwargsParser,), {})

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__lowercase(self):
        victim = self.Victim("")
        assert victim.SOURCE == ""
        assert victim.PREFIX == ""
        assert victim.CMD == ""
        assert victim.ARGS == []
        assert victim.KWARGS == {}

        victim = self.Victim("hello")
        assert victim.SOURCE == "hello"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == []
        assert victim.KWARGS == {}

        victim = self.Victim("HELLO")
        assert victim.SOURCE == "HELLO"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == []
        assert victim.KWARGS == {}

    def test__args_kwargs(self):
        victim = self.Victim("HELLO CH")
        assert victim.SOURCE == "HELLO CH"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == ["ch", ]
        assert victim.KWARGS == {}

        victim = self.Victim("HELLO CH 1")
        assert victim.SOURCE == "HELLO CH 1"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == ["ch", "1", ]
        assert victim.KWARGS == {}

        victim = self.Victim("HELLO CH=1")
        assert victim.SOURCE == "HELLO CH=1"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1"}

        victim = self.Victim("HELLO CH1 CH2=2    ch3=3  ch4")
        assert victim.SOURCE == "HELLO CH1 CH2=2    ch3=3  ch4"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == ["ch1", "ch4"]
        assert victim.KWARGS == {"ch2": "2", "ch3": "3"}

    def test__kwargs_spaces(self):
        victim = self.Victim("CH =  1")
        assert victim.SOURCE == "CH =  1"
        assert victim.PREFIX == ""
        assert victim.CMD == ""
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1"}

        victim = self.Victim("hello CH =  1")
        assert victim.SOURCE == "hello CH =  1"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1"}

        victim = self.Victim("hello CH ===  1")
        assert victim.SOURCE == "hello CH ===  1"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1"}

    def test__kwargs_only(self):
        victim = self.Victim("CH=1")
        assert victim.SOURCE == "CH=1"
        assert victim.PREFIX == ""
        assert victim.CMD == ""
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1"}

        victim = self.Victim("CH=1 ch2=2")
        assert victim.SOURCE == "CH=1 ch2=2"
        assert victim.PREFIX == ""
        assert victim.CMD == ""
        assert victim.ARGS == []
        assert victim.KWARGS == {"ch": "1", "ch2": "2"}

    def test__prefix(self):
        victim = self.Victim("HELLO CH 1", prefix_expected="HELLO")
        assert victim.SOURCE == "HELLO CH 1"
        assert victim.PREFIX == "hello"
        assert victim.CMD == "ch"
        assert victim.ARGS == ["1", ]
        assert victim.KWARGS == {}

        victim = self.Victim("HELLO CH 1", prefix_expected="HELLO CH 1")
        assert victim.SOURCE == "HELLO CH 1"
        assert victim.PREFIX == "hello ch 1"
        assert victim.CMD == ""
        assert victim.ARGS == []
        assert victim.KWARGS == {}

        victim = self.Victim("HELLO CH 1", prefix_expected="HELLO123")
        assert victim.SOURCE == "HELLO CH 1"
        assert victim.PREFIX == ""
        assert victim.CMD == "hello"
        assert victim.ARGS == ["ch", "1", ]
        assert victim.KWARGS == {}


# =====================================================================================================================



from typing import *
import os
import re

from base_aux.aux_attr.m1_annot_attr1_aux import AttrAux_AnnotsAll
from base_aux.aux_attr.m4_kits import AttrKit_Blank
from base_aux.base_types.m1_type_aux import TypeAux
from base_aux.base_nest_dunders.m1_init2_annots1_attrs_by_args_kwargs import NestInit_AnnotsAttr_ByArgsKwargs
from base_aux.base_nest_dunders.m3_calls import NestCall_Resolve


# =====================================================================================================================
class PvLoaderEnv(NestCall_Resolve):
    # INIT -------
    TARGET: type[NestInit_AnnotsAttr_ByArgsKwargs] | Any = AttrKit_Blank
    PATTS: tuple[str, ...] = ()

    def __init__(
            self,
            target: type | Any = None,
            patts: tuple[str, ...] = None,
            **kwargs,
    ) -> None | NoReturn:
        super().__init__(**kwargs)

        if target is not None:
            self.TARGET = target

        if patts is not None:
            self.PATTS = patts

    # -----------------------------------------------------------------------------------------------------------------
    def resolve(self) -> NestInit_AnnotsAttr_ByArgsKwargs | Any | NoReturn:
        # get dict -------
        data = dict(os.environ)     # just a copy!

        # filter ---
        if self.PATTS:
            filtered_out = filter(lambda name: not any([re.search(pat, name, flags=re.IGNORECASE) for pat in self.PATTS]), list(data))
            for out_i in filtered_out:
                data.pop(out_i)

        # load args -------
        if TypeAux(self.TARGET).check__class() and issubclass(self.TARGET, NestInit_AnnotsAttr_ByArgsKwargs):
            # used for check Annots all inited!

            result = self.TARGET(**data)
        else:
            AttrAux_AnnotsAll(self.TARGET).sai__by_args_kwargs(**data)
            result = self.TARGET

        return result


# =====================================================================================================================



# DEEP SEEK

import os
from typing import List, Dict, Optional, Tuple
from datetime import datetime
try:
    from git import Repo, GitCommandError, InvalidGitRepositoryError
    GITPYTHON_AVAILABLE = True
except ImportError:
    GITPYTHON_AVAILABLE = False


# =====================================================================================================================
class GitPythonManager:
    def __init__(self, repo_path: str):
        if not GITPYTHON_AVAILABLE:
            raise ImportError("GitPython не установлен. Установите: pip install GitPython")
        
        self.repo_path = os.path.abspath(repo_path)
        try:
            self.repo = Repo(self.repo_path)
        except InvalidGitRepositoryError:
            raise ValueError(f"Путь {self.repo_path} не является валидным Git репозиторием")
    
    # def get_branches(self) -> List[Dict[str, str]]:
    #     """Получает список всех веток проекта"""
    #     branches = []
    #
    #     # Локальные ветки
    #     for branch in self.repo.branches:
    #         branches.append({
    #             'name': branch.name,
    #             'type': 'local',
    #             'hash': branch.commit.hexsha
    #         })
    #
    #     # Удаленные ветки
    #     for remote in self.repo.remotes:
    #         for ref in remote.refs:
    #             branches.append({
    #                 'name': ref.name,
    #                 'type': 'remote',
    #                 'hash': ref.commit.hexsha if hasattr(ref, 'commit') else 'unknown'
    #             })
    #
    #     return branches
    
    def get_commits(self, branch_name: str, limit: int = 10) -> list[dict[str, str]]:
        """Получает коммиты указанной ветки с ограничением количества"""
        try:
            commits = []
            for commit in self.repo.iter_commits(branch_name, max_count=limit):
                commits.append({
                    'hash': commit.hexsha,
                    'date': commit.committed_datetime.isoformat(),
                    'author': commit.author.name,
                    'email': commit.author.email,
                    'message': commit.message.strip()
                })
            return commits
        except GitCommandError:
            return []
    
    def get_current_state(self) -> Dict[str, str]:
        """Получает текущее состояние репозитория"""
        state = {}
        
        # Текущая ветка
        try:
            state['current_branch'] = self.repo.active_branch.name
        except TypeError:
            state['current_branch'] = 'DETACHED_HEAD'
        
        # Текущий коммит
        state['current_hash'] = self.repo.head.commit.hexsha
        state['current_date'] = self.repo.head.commit.committed_datetime.isoformat()
        state['current_author'] = self.repo.head.commit.author.name
        
        # Информация о проекте
        state['project_name'] = os.path.basename(self.repo_path)
        state['repo_path'] = self.repo_path
        
        # Состояние репозитория
        state['is_valid'] = not self.repo.bare
        state['has_changes'] = self.repo.is_dirty()
        
        # Детальная информация о изменениях
        state['status_details'] = self._get_detailed_status()
        
        return state
    
    def _get_detailed_status(self) -> Dict[str, List[str]]:
        """Получает детальную информацию о состоянии файлов"""
        changed_files = [item.a_path for item in self.repo.index.diff(None)]  # Неиндексированные
        staged_files = [item.a_path for item in self.repo.index.diff('HEAD')]  # Индексированные
        untracked_files = self.repo.untracked_files
        
        return {
            'staged': staged_files,
            'unstaged': changed_files,
            'untracked': untracked_files
        }
    
    def switch_to_commit(self, branch_name: str, commit_hash: Optional[str] = None) -> Tuple[bool, str]:
        """Переключается на указанную ветку и коммит"""
        try:
            # Переключаемся на ветку
            self.repo.git.checkout(branch_name)
            
            # Если указан коммит, переключаемся на него
            if commit_hash:
                self.repo.git.checkout(commit_hash)
            
            return True, "Успешное переключение"
        except GitCommandError as e:
            return False, f"Ошибка переключения: {str(e)}"


# =====================================================================================================================
# Универсальный менеджер с выбором реализации
class UniversalGitManager:
    def __init__(self, repo_path: str, use_gitpython: bool = True):
        """
        Универсальный менеджер Git репозиториев
        
        :param repo_path: путь к репозиторию
        :param use_gitpython: использовать GitPython если доступен
        """
        self.use_gitpython = use_gitpython and GITPYTHON_AVAILABLE
        
        if self.use_gitpython:
            self.manager = GitPythonManager(repo_path)
            self.impl_type = "GitPython"
        else:
            # Здесь можно добавить реализацию через subprocess
            self.impl_type = "subprocess"
            raise NotImplementedError("Реализация через subprocess не добавлена")
    
    def get_implementation_type(self) -> str:
        return self.impl_type
    
    # Прокси-методы к выбранной реализации
    def get_branches(self):
        return self.manager.get_branches()
    
    def get_commits(self, branch_name, limit=10):
        return self.manager.get_commits(branch_name, limit)
    
    def get_current_state(self):
        return self.manager.get_current_state()
    
    def switch_to_commit(self, branch_name, commit_hash=None):
        return self.manager.switch_to_commit(branch_name, commit_hash)

# Пример использования
if __name__ == "__main__":
    repo_path = input("Введите путь к Git репозиторию: ").strip()
    
    try:
        # Автоматический выбор лучшей реализации
        manager = UniversalGitManager(repo_path)
        print(f"✓ Используется реализация: {manager.get_implementation_type()}")
        
        # Тестирование функционала
        print("\nТекущее состояние:")
        state = manager.get_current_state()
        print(f"Ветка: {state['current_branch']}")
        print(f"Коммит: {state['current_hash'][:8]}")
        
        print("\nВетки:")
        for branch in manager.get_branches()[:5]:  # Первые 5 веток
            print(f"  - {branch['name']}")
        
    except Exception as e:
        print(f"Ошибка: {e}")